# Copyright 2016-2018 Berk Onat, Fawzi Mohamed
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import argparse

def get_commandLineArguments(cmdline):
    """Parse command line arguments of GROMACS
       using argparse

    Returns:
        the dictionary of arguments
    """
    cmdlineparser = argparse.ArgumentParser()
    cmdlineparser.add_argument('-s', nargs='?', const='topol.tpr', default='topol.tpr',
        help='[<.tpr/.tpb/...>](Input)  Run input file: tpr tpb tpa')
    cmdlineparser.add_argument('-o', nargs='?', const='traj.trr', default='traj.trr',
        help='[<.trr/.cpt/...>](Output)  Full precision trajectory: trr cpt trj tng')
    cmdlineparser.add_argument('-x', nargs='?', const='traj_comp.xtc', default='traj_comp.xtc',
        help='[<.xtc/.tng>](Output, Optional) Compressed trajectory (tng format or portable xdr format)')
    cmdlineparser.add_argument('-cpi', nargs='?', const='state.cpt', default='state.cpt',
        help='[<.cpt>](Input, Optional) Checkpoint file')
    cmdlineparser.add_argument('-cpo', nargs='?', const='state.cpt', default='state.cpt',
        help='[<.cpt>](Output, Optional) Checkpoint file')
    cmdlineparser.add_argument('-c', nargs='?', const='confout.gro', default='confout.gro',
        help='[<.gro/.g96/...>](Output)  Structure file: gro g96 pdb brk ent esp')
    cmdlineparser.add_argument('-e', nargs='?', const='ener.edr', default='ener.edr',
        help='[<.edr>](Output)  Energy file')
    cmdlineparser.add_argument('-g', nargs='?', const='md.log', default='md.log',
        help='[<.log>](Output)  Log file')
    cmdlineparser.add_argument('-dhdl', nargs='?', const='dhdl.xvg', default='dhdl.xvg',
        help='[<.xvg>](Output, Optional) xvgr/xmgr file')
    cmdlineparser.add_argument('-field', nargs='?', const='field.xvg', default='field.xvg',
        help='[<.xvg>](Output, Optional) xvgr/xmgr file')
    cmdlineparser.add_argument('-table', nargs='?', const='table.xvg', default='table.xvg',
        help='[<.xvg>](Input, Optional) xvgr/xmgr file')
    cmdlineparser.add_argument('-tabletf', nargs='?', const='tabletf.xvg', default='tabletf.xvg',
        help='[<.xvg>](Input, Optional) xvgr/xmgr file')
    cmdlineparser.add_argument('-tablep', nargs='?', const='tablep.xvg', default='tablep.xvg',
        help='[<.xvg>](Input, Optional) xvgr/xmgr file')
    cmdlineparser.add_argument('-tableb', nargs='?', const='table.xvg', default='table.xvg',
        help='[<.xvg>](Input, Optional) xvgr/xmgr file')
    cmdlineparser.add_argument('-rerun', nargs='?', const='rerun.xtc', default='rerun.xtc',
        help='[<.xtc/.trr/...>](Input, Optional) Trajectory: xtc trr cpt trj gro g96 pdb tng')
    cmdlineparser.add_argument('-tpi', nargs='?', const='tpi.xvg', default='tpi.xvg',
        help='[<.xvg>](Output, Optional) xvgr/xmgr file')
    cmdlineparser.add_argument('-tpid', nargs='?', const='tpidist.xvg', default='tpidist.xvg',
        help='[<.xvg>](Output, Optional) xvgr/xmgr file')
    cmdlineparser.add_argument('-ei', nargs='?', const='sam.edi', default='sam.edi',
        help='[<.edi>](Input, Optional) ED sampling input')
    cmdlineparser.add_argument('-eo', nargs='?', const='edsam.xvg', default='edsam.xvg',
        help='[<.xvg>](Output, Optional) xvgr/xmgr file')
    cmdlineparser.add_argument('-devout', nargs='?', const='deviatie.xvg', default='deviatie.xvg',
        help='[<.xvg>](Output, Optional) xvgr/xmgr file')
    cmdlineparser.add_argument('-runav', nargs='?', const='runaver.xvg', default='runaver.xvg',
        help='[<.xvg>](Output, Optional) xvgr/xmgr file')
    cmdlineparser.add_argument('-px', nargs='?', const='pullx.xvg', default='pullx.xvg',
        help='[<.xvg>](Output, Optional) xvgr/xmgr file')
    cmdlineparser.add_argument('-pf', nargs='?', const='pullf.xvg', default='pullf.xvg',
        help='[<.xvg>](Output, Optional) xvgr/xmgr file')
    cmdlineparser.add_argument('-ro', nargs='?', const='rotation.xvg', default='rotation.xvg',
        help='[<.xvg>](Output, Optional) xvgr/xmgr file')
    cmdlineparser.add_argument('-ra', nargs='?', const='rotangles.log', default='rotangles.log',
        help='[<.log>](Output, Optional) Log file')
    cmdlineparser.add_argument('-rs', nargs='?', const='rotslabs.log', default='rotslabs.log',
        help='[<.log>](Output, Optional) Log file')
    cmdlineparser.add_argument('-rt', nargs='?', const='rottorque.log', default='rottorque.log',
        help='[<.log>](Output, Optional) Log file')
    cmdlineparser.add_argument('-mtx', nargs='?', const='nm.mtx', default='nm.mtx',
        help='[<.mtx>](Output, Optional) Hessian matrix')
    cmdlineparser.add_argument('-dn', nargs='?', const='dipole.ndx', default='dipole.ndx',
        help='[<.ndx>](Output, Optional) Index file')
    cmdlineparser.add_argument('-multidir', nargs='?', const='rundir', default='rundir',
        help='[<dir> [...]](Input, Optional) Run directory')
    cmdlineparser.add_argument('-membed', nargs='?', const='membed.dat', default='membed.dat',
        help='[<.dat>](Input, Optional) Generic data file')
    cmdlineparser.add_argument('-mp', nargs='?', const='membed.top', default='membed.top',
        help='[<.top>](Input, Optional) Topology file')
    cmdlineparser.add_argument('-mn', nargs='?', const='membed.ndx', default='membed.ndx',
        help='[<.ndx>](Input, Optional) Index file')
    cmdlineparser.add_argument('-if', nargs='?', const='imdforces.xvg', default='imdforces.xvg',
        help='[<.xvg>](Output, Optional) xvgr/xmgr file')
    cmdlineparser.add_argument('-swap', nargs='?', const='swapions.xvg', default='swapions.xvg',
        help='[<.xvg>](Output, Optional) xvgr/xmgr file')
    cmdlineparser.add_argument('-nice', nargs='?', const='0', default='0',
        help='<int>  Set the nicelevel')
    cmdlineparser.add_argument('-deffnm', nargs='?', const='', default='',
        help='<string>  Set the default filename for all file options')
    cmdlineparser.add_argument('-xvg', nargs='?', const='xmgrace', default='xmgrace',
        help='<enum>  xvg plot formatting: xmgrace, xmgr, none')
    cmdlineparser.add_argument('-dd', nargs='?', const='0 0 0', default='0 0 0',
        help='<vector> Domain decomposition grid, 0 is optimize')
    cmdlineparser.add_argument('-ddorder', nargs='?', const='interleave', default='interleave',
        help='<enum>  DD rank order: interleave, pp_pme, cartesian')
    cmdlineparser.add_argument('-npme', nargs='?', const='-1', default='-1',
        help='<int>  Number of separate ranks to be used for PME, -1 is guess')
    cmdlineparser.add_argument('-nt', nargs='?', const='0', default='0',
        help='<int>  Total number of threads to start (0 is guess)')
    cmdlineparser.add_argument('-ntmpi', nargs='?', const='0', default='0',
        help='<int>  Number of thread-MPI threads to start (0 is guess)')
    cmdlineparser.add_argument('-ntomp', nargs='?', const='0', default='0',
        help='<int>  Number of OpenMP threads per MPI rank to start (0 is guess)')
    cmdlineparser.add_argument('-ntomp_pme', nargs='?', const='0', default='0',
        help='<int>  Number of OpenMP threads per MPI rank to start (0 is -ntomp)')
    cmdlineparser.add_argument('-pin', nargs='?', const='auto', default='auto',
        help='<enum>  Set thread affinities: auto, on, off')
    cmdlineparser.add_argument('-pinoffset', nargs='?', const='0', default='0',
        help='<int>  The starting logical core number for pinning to cores; used to avoid pinning threads from different mdrun instances to the same core')
    cmdlineparser.add_argument('-pinstride', nargs='?', const='0', default='0',
        help='<int>  Pinning distance in logical cores for threads, use 0 to minimize the number of threads per physical core')
    cmdlineparser.add_argument('-gpu_id', nargs='?', const='', default='',
        help='<string>  List of GPU device id-s to use, specifies the per-node PP rank to GPU mapping')
    cmdlineparser.add_argument('-ddcheck', nargs='?', const=True, default='yes',
        help='  Check for all bonded interactions with DD')
    cmdlineparser.add_argument('-noddcheck', nargs='?', const=False, default='yes',
        help='  Check for all bonded interactions with DD')
    cmdlineparser.add_argument('-rdd', nargs='?', const='0', default='0',
        help='<real>  The maximum distance for bonded interactions with DD (nm), 0 is determine from initial coordinates')
    cmdlineparser.add_argument('-rcon', nargs='?', const='0', default='0',
        help='<real>  Maximum distance for P-LINCS (nm), 0 is estimate')
    cmdlineparser.add_argument('-dlb', nargs='?', const='auto', default='auto',
        help='<enum>  Dynamic load balancing (with DD): auto, no, yes')
    cmdlineparser.add_argument('-dds', nargs='?', const='0.8', default='0.8',
        help='<real>  Fraction in (0,1) by whose reciprocal the initial DD cell size will be increased in order to provide a margin in which dynamic load balancing can act while preserving the minimum cell size.')
    cmdlineparser.add_argument('-gcom', nargs='?', const='-1', default='-1',
        help='<int>  Global communication frequency')
    cmdlineparser.add_argument('-nb', nargs='?', const='auto', default='auto',
        help='<enum>  Calculate non-bonded interactions on: auto, cpu, gpu, gpu_cpu')
    cmdlineparser.add_argument('-nstlist', nargs='?', const='0', default='0',
        help='<int>  Set nstlist when using a Verlet buffer tolerance (0 is guess)')
    cmdlineparser.add_argument('-tunepme', nargs='?', const=True, default='yes',
        help='  Optimize PME load between PP/PME ranks or GPU/CPU')
    cmdlineparser.add_argument('-notunepme', nargs='?', const=False, default='yes',
        help='  Optimize PME load between PP/PME ranks or GPU/CPU')
    cmdlineparser.add_argument('-testverlet', nargs='?', const=True, default='no',
        help='  Test the Verlet non-bonded scheme')
    cmdlineparser.add_argument('-notestverlet', nargs='?', const=False, default='no',
        help='  Test the Verlet non-bonded scheme')
    cmdlineparser.add_argument('-v', nargs='?', const=True, default='no',
        help='  Be loud and noisy')
    cmdlineparser.add_argument('-nov', nargs='?', const=False, default='no',
        help='  Be loud and noisy')
    cmdlineparser.add_argument('-compact', nargs='?', const=True, default='yes',
        help='  Write a compact log file')
    cmdlineparser.add_argument('-nocompact', nargs='?', const=False, default='yes',
        help='  Write a compact log file')
    cmdlineparser.add_argument('-seppot', nargs='?', const=True, default='no',
        help='  Write separate V and dVdl terms for each interaction type and rank to the log file(s)')
    cmdlineparser.add_argument('-noseppot', nargs='?', const=False, default='no',
        help='  Write separate V and dVdl terms for each interaction type and rank to the log file(s)')
    cmdlineparser.add_argument('-pforce', nargs='?', const='-1', default='-1',
        help='<real>  Print all forces larger than this (kJ/mol nm)')
    cmdlineparser.add_argument('-reprod', nargs='?', const=True, default='no',
        help='  Try to avoid optimizations that affect binary reproducibility')
    cmdlineparser.add_argument('-noreprod', nargs='?', const=False, default='no',
        help='  Try to avoid optimizations that affect binary reproducibility')
    cmdlineparser.add_argument('-cpt', nargs='?', const='15', default='15',
        help='<real>  Checkpoint interval (minutes)')
    cmdlineparser.add_argument('-cpnum', nargs='?', const=True, default='no',
        help='  Keep and number checkpoint files')
    cmdlineparser.add_argument('-nocpnum', nargs='?', const=False, default='no',
        help='  Keep and number checkpoint files')
    cmdlineparser.add_argument('-append', nargs='?', const=True, default='yes',
        help='  Append to previous output files when continuing from checkpoint instead of adding the simulation part number to all file names')
    cmdlineparser.add_argument('-noappend', nargs='?', const=False, default='yes',
        help='  Append to previous output files when continuing from checkpoint instead of adding the simulation part number to all file names')
    cmdlineparser.add_argument('-nsteps', nargs='?', const='-2', default='-2',
        help='<int>  Run this number of steps, overrides .mdp file option')
    cmdlineparser.add_argument('-maxh', nargs='?', const='-1', default='-1',
        help='<real>  Terminate after 0.99 times this time (hours)')
    cmdlineparser.add_argument('-multi', nargs='?', const='0', default='0',
        help='<int>  Do multiple simulations in parallel')
    cmdlineparser.add_argument('-replex', nargs='?', const='0', default='0',
        help='<int>  Attempt replica exchange periodically with this period (steps)')
    cmdlineparser.add_argument('-nex', nargs='?', const='0', default='0',
        help='<int>  Number of random exchanges to carry out each exchange interval (N^3 is one suggestion). -nex zero or not specified gives neighbor replica exchange.')
    cmdlineparser.add_argument('-reseed', nargs='?', const='-1', default='-1',
        help='<int>  Seed for replica exchange, -1 is generate a seed')
    args = cmdlineparser.parse_args(str(cmdline).split())
    return vars(args)




# Copyright 2016-2018 The NOMAD Developers Group
# Copyright 2017-2018 Lorenzo Pardini
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import xml.sax
import logging
import numpy as np
from nomadcore.unit_conversion.unit_conversion import convert_unit_function
from nomadcore.unit_conversion.unit_conversion import convert_unit
from nomadcore.unit_conversion import unit_conversion


class InputHandler(xml.sax.handler.ContentHandler):
    def __init__(self, backend, gmaxvr):
        self.gmaxvr = gmaxvr
        self.backend = backend
        self.inputSectionGIndex = -1
        self.inGWInput = False
        self.freqmax = 1.0
        self.singularity = 'mpd'
        self.actype = 'pade'
        self.npol = 0
        self.scrtype = "rpa"
        self.snempty = 0
        self.pnempty = 0
        self.coreflag = "all"
        self.fgrid = "gaule2"
        self.lmaxmb = 3
        self.epsmb = 0.0001
        self.gmb = 1.0
        self.sciavtype = "isotropic"
        self.cutofftype = "none"
        self.pwm = 2.0
        self.ngridqDum = [1, 1, 1]
        self.ngridq = [1, 1, 1]
        self.nomeg = 16

        self.freqgrid = "none"
        self.selfenergy = "none"
        self.mixbasis = "none"
        self.barecoul = "none"
        self.scrcoul = "none"

    def endDocument(self):
        if self.freqgrid == "none":
            self.backend.addValue("gw_max_frequency", self.freqmax)
            self.backend.addValue("gw_frequency_grid_type", self.fgrid)
            self.backend.addValue("gw_number_of_frequencies", self.nomeg)
        if self.selfenergy == "none":
            self.backend.addValue("gw_self_energy_c_number_of_poles", int(self.npol))
            self.backend.addValue("gw_self_energy_c_number_of_empty_states", int(self.snempty))
            self.backend.addValue("gw_self_energy_singularity_treatment", self.singularity)
            self.backend.addValue("gw_self_energy_c_analytical_continuation", self.actype)
        if self.mixbasis == "none":
            self.backend.addValue("gw_mixed_basis_lmax", self.lmaxmb)
            self.backend.addValue("gw_mixed_basis_tolerance", self.epsmb)
            self.backend.addValue("gw_mixed_basis_gmax", self.gmb*self.gmaxvr)
        if self.barecoul == "none":
            self.backend.addValue("gw_bare_coulomb_gmax", self.pwm*self.gmb*self.gmaxvr)
            self.backend.addValue("gw_bare_coulomb_cutofftype", self.cutofftype)
        if self.scrcoul == "none":
            self.backend.addValue("gw_screened_coulomb_volume_average",self.sciavtype)
            self.backend.addValue("gw_screened_Coulomb", self.scrtype)
        self.backend.addValue("gw_basis_set", "mixed")
        self.backend.addValue("gw_qp_equation_treatment", "linearization")
        for j in range(0, 3):
            # This is causing an error: Invalid literal for int()
            # with base 10.
            self.ngridq[j] = int(self.ngridqDum[j])
        self.backend.addValue("gw_ngridq", self.ngridq)

    def startElement(self, name, attrs):
        fromH = unit_conversion.convert_unit_function("hartree", "J")
        if name == "gw":
            # self.inputSectionGIndex = self.backend.openSection("section_system")
            self.inGWInput = True
            try:
                self.coreflag = attrs.getValue('coreflag')
                self.backend.addValue("gw_core_treatment", self.coreflag)
            except:
                self.coreflag = "all"
                self.backend.addValue("gw_core_treatment", self.coreflag)
            try:
                self.pnempty = attrs.getValue('nempty')
                self.backend.addValue("gw_polarizability_number_of_empty_states", int(self.pnempty))
            except:
                self.pnempty = 0
                self.backend.addValue("gw_polarizability_number_of_empty_states", int(self.pnempty))
            try:
                dummy = attrs.getValue('ngridq')
                self.ngridqDum = dummy.split()
            except:
                self.ngridqDum = [1, 1, 1]
        elif name == "freqgrid":
            self.freqgrid = "freqgrid"
            try:
                self.freqmax = attrs.getValue('freqmax')
                self.backend.addValue("gw_max_frequency", float(self.freqmax))
            except:
                self.freqmax = 1.0
                self.backend.addValue("gw_max_frequency", self.freqmax)
            try:
                self.fgrid = attrs.getValue('fgrid')
                self.backend.addValue("gw_frequency_grid_type", self.fgrid)
            except:
                self.fgrid = "gaule2"
                self.backend.addValue("gw_frequency_grid_type", self.fgrid)
            try:
                self.nomeg = attrs.getValue('nomeg')
                self.backend.addValue("gw_number_of_frequencies", int(self.nomeg))
            except:
                self.nomeg = 16
                self.backend.addValue("gw_number_of_frequencies", self.nomeg)

        elif name == "selfenergy":
            self.selfenergy = "selfenergy"
            try:
                self.npol = attrs.getValue('npol')
                self.backend.addValue("gw_self_energy_c_number_of_poles", int(self.npol))
            except:
                self.npol = 0
                self.backend.addValue("gw_self_energy_c_number_of_poles", self.npol)
            try:
                self.snempty = attrs.getValue('nempty')
                self.backend.addValue("gw_self_energy_c_number_of_empty_states", int(self.snempty))
            except:
                self.snempty = 0
                self.backend.addValue("gw_self_energy_c_number_of_empty_states", self.snempty)
            try:
                self.singularity = attrs.getValue('singularity')
                self.backend.addValue("gw_self_energy_singularity_treatment", self.singularity)
            except:
                self.singularity = 'mpd'
                self.backend.addValue("gw_self_energy_singularity_treatment", self.singularity)
            try:
                self.actype = attrs.getValue('actype')
                self.backend.addValue("gw_self_energy_c_analytical_continuation", self.actype)
            except:
                self.actype = 'pade'
                self.backend.addValue("gw_self_energy_c_analytical_continuation", self.actype)

        elif name == "mixbasis":
            self.mixbasis = "mixbasis"
            try:
                self.lmaxmb = attrs.getValue('lmaxmb')
                self.backend.addValue("gw_mixed_basis_lmax", int(self.lmaxmb))
            except:
                self.lmaxmb = 3
                self.backend.addValue("gw_mixed_basis_lmax", self.lmaxmb)
            try:
                self.epsmb = attrs.getValue('epsmb')
                self.backend.addValue("gw_mixed_basis_tolerance", float(self.epsmb))
            except:
                self.epsmb = 0.0001
                self.backend.addValue("gw_mixed_basis_tolerance", self.epsmb)
            try:
                self.gmb = attrs.getValue('gmb')
                self.backend.addValue("gw_mixed_basis_gmax", float(self.gmb)*self.gmaxvr)
            except:
                self.gmb = 1.0
                self.backend.addValue("gw_mixed_basis_gmax", self.gmb*self.gmaxvr)

        elif name == "barecoul":
            self.barecoul = "barecoul"
            try:
                self.pwm = attrs.getValue('pwm')
                self.backend.addValue("gw_bare_coulomb_gmax", float(self.pwm)*float(self.gmb)*self.gmaxvr)
            except:
                self.pwm = 2.0
                self.backend.addValue("gw_bare_coulomb_gmax", self.pwm*float(self.gmb)*self.gmaxvr)
            try:
                self.cutofftype = attrs.getValue('cutofftype')
                self.backend.addValue("gw_bare_coulomb_cutofftype", self.cutofftype)
            except:
                self.cutofftype = "none"
                self.backend.addValue("gw_bare_coulomb_cutofftype", self.cutofftype)

        elif name == "scrcoul":
            self.scrcoul = "scrcoul"
            try:
                self.sciavtype = attrs.getValue('sciavtype')
                self.backend.addValue("gw_screened_coulomb_volume_average",self.sciavtype)
            except:
                self.sciavtype = "isotropic"
                self.backend.addValue("gw_screened_coulomb_volume_average",self.sciavtype)
            try:
                self.scrtype = attrs.getValue('scrtype')
                self.backend.addValue("gw_screened_Coulomb", self.scrtype)
            except:
                self.scrtype = "rpa"
                self.backend.addValue("gw_screened_Coulomb", self.scrtype)

    def endElement(self, name):
        pass

def parseInput(inF, backend, gmaxvr):
    handler = InputHandler(backend, gmaxvr)
    logging.info("will parse")
    xml.sax.parse(inF, handler)
    logging.info("did parse")



# pylint: disable=line-too-long
import json
import os
from pathlib import Path

from lark.exceptions import VisitError

from pydantic import ValidationError
from fastapi import FastAPI
from fastapi.exceptions import RequestValidationError, StarletteHTTPException
from fastapi.middleware.cors import CORSMiddleware

from optimade import __api_version__, __version__
import optimade.server.exception_handlers as exc_handlers

from .entry_collections import MongoCollection
from .config import CONFIG
from .middleware import EnsureQueryParamIntegrity
from .routers import info, links, references, structures, landing
from .routers.utils import get_providers, BASE_URL_PREFIXES


if CONFIG.debug:  # pragma: no cover
    print("DEBUG MODE")


app = FastAPI(
    title="OPTIMADE API",
    description=(
        f"""The [Open Databases Integration for Materials Design (OPTIMADE) consortium](https://www.optimade.org/) aims to make materials databases interoperational by developing a common REST API.

This specification is generated using [`optimade-python-tools`](https://github.com/Materials-Consortia/optimade-python-tools/tree/v{__version__}) v{__version__}."""
    ),
    version=__api_version__,
    docs_url=f"{BASE_URL_PREFIXES['major']}/extensions/docs",
    redoc_url=f"{BASE_URL_PREFIXES['major']}/extensions/redoc",
    openapi_url=f"{BASE_URL_PREFIXES['major']}/extensions/openapi.json",
)


if not CONFIG.use_real_mongo:
    import bson.json_util
    import optimade.server.data as data
    from .routers import ENTRY_COLLECTIONS

    def load_entries(endpoint_name: str, endpoint_collection: MongoCollection):
        print(f"loading test {endpoint_name}...")

        endpoint_collection.collection.insert_many(getattr(data, endpoint_name, []))
        if endpoint_name == "links":
            print("adding Materials-Consortia providers to links from optimade.org")
            endpoint_collection.collection.insert_many(
                bson.json_util.loads(bson.json_util.dumps(get_providers()))
            )
        print(f"done inserting test {endpoint_name}...")

    for name, collection in ENTRY_COLLECTIONS.items():
        load_entries(name, collection)


# Add various middleware
app.add_middleware(CORSMiddleware, allow_origins=["*"])
app.add_middleware(EnsureQueryParamIntegrity)


# Add various exception handlers
app.add_exception_handler(StarletteHTTPException, exc_handlers.http_exception_handler)
app.add_exception_handler(
    RequestValidationError, exc_handlers.request_validation_exception_handler
)
app.add_exception_handler(ValidationError, exc_handlers.validation_exception_handler)
app.add_exception_handler(VisitError, exc_handlers.grammar_not_implemented_handler)
app.add_exception_handler(NotImplementedError, exc_handlers.not_implemented_handler)
app.add_exception_handler(Exception, exc_handlers.general_exception_handler)


# Add various endpoints to `/vMAJOR`
app.include_router(info.router, prefix=BASE_URL_PREFIXES["major"])
app.include_router(links.router, prefix=BASE_URL_PREFIXES["major"])
app.include_router(references.router, prefix=BASE_URL_PREFIXES["major"])
app.include_router(structures.router, prefix=BASE_URL_PREFIXES["major"])


# Add the router for the landing page for all prefixes
app.include_router(landing.router)
app.include_router(landing.router, prefix=BASE_URL_PREFIXES["major"])


def add_optional_versioned_base_urls(app: FastAPI):
    """Add the following OPTIONAL prefixes/base URLs to server:
    ```
        /vMajor.Minor
        /vMajor.Minor.Patch
    ```
    """
    for version in ("minor", "patch"):
        app.include_router(info.router, prefix=BASE_URL_PREFIXES[version])
        app.include_router(links.router, prefix=BASE_URL_PREFIXES[version])
        app.include_router(references.router, prefix=BASE_URL_PREFIXES[version])
        app.include_router(structures.router, prefix=BASE_URL_PREFIXES[version])
        app.include_router(landing.router, prefix=BASE_URL_PREFIXES[version])


def update_schema(app: FastAPI):
    """Update OpenAPI schema in file 'local_openapi.json'"""
    package_root = Path(__file__).parent.parent.parent.resolve()
    if not package_root.joinpath("openapi").exists():
        os.mkdir(package_root.joinpath("openapi"))
    with open(package_root.joinpath("openapi/local_openapi.json"), "w") as f:
        json.dump(app.openapi(), f, indent=2)


@app.on_event("startup")
async def startup_event():
    # Update OpenAPI schema on versioned base URL `/vMAJOR`
    update_schema(app)
    # Add API endpoints for OPTIONAL base URLs `/vMAJOR.MINOR` and `/vMAJOR.MINOR.PATCH`
    add_optional_versioned_base_urls(app)



# pylint: disable=no-self-argument
import json
import logging
from typing import Optional, Dict, List

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal
from pathlib import Path

from pydantic import BaseSettings, Field, root_validator

from optimade import __version__
from optimade.models import Implementation, Provider


DEFAULT_CONFIG_FILE_PATH = str(Path.home().joinpath(".optimade.json"))
logger = logging.getLogger("optimade")


class NoFallback(Exception):
    """No fallback value can be found."""


class ServerConfig(BaseSettings):
    """ This class stores server config parameters in a way that
    can be easily extended for new config file types.

    """

    config_file: str = Field(
        DEFAULT_CONFIG_FILE_PATH, description="File to load alternative defaults from",
    )
    debug: bool = Field(
        False, description="Turns on Debug Mode for the OPTIMADE Server implementation"
    )
    use_real_mongo: bool = Field(
        False, description="Use a real Mongo server rather than MongoMock"
    )
    mongo_database: str = Field(
        "optimade", description="Mongo database for collection data"
    )
    mongo_uri: str = Field("localhost:27017", description="URI for the Mongo server")
    links_collection: str = Field(
        "links", description="Mongo collection name for /links endpoint resources"
    )
    references_collection: str = Field(
        "references",
        description="Mongo collection name for /references endpoint resources",
    )
    structures_collection: str = Field(
        "structures",
        description="Mongo collection name for /structures endpoint resources",
    )
    page_limit: int = Field(20, description="Default number of resources per page")
    page_limit_max: int = Field(
        500, description="Max allowed number of resources per page"
    )
    default_db: str = Field(
        "test_server",
        description="ID of /links endpoint resource for the chosen default OPTIMADE implementation (only relevant for the index meta-database)",
    )
    base_url: Optional[str] = Field(
        None, description="Base URL for this implementation"
    )
    implementation: Implementation = Field(
        Implementation(
            name="Example implementation",
            version=__version__,
            source_url="https://github.com/Materials-Consortia/optimade-python-tools",
            maintainer=None,
        ),
        description="Introspective information about this OPTIMADE implementation",
    )
    provider: Provider = Field(
        Provider(
            prefix="exmpl",
            name="Example provider",
            description="Provider used for examples, not to be assigned to a real database",
            homepage="https://example.com",
            index_base_url="http://localhost:5001",
        ),
        description="General information about the provider of this OPTIMADE implementation",
    )
    provider_fields: Dict[
        Literal["links", "references", "structures"], List[str]
    ] = Field(
        {},
        description="A list of additional fields to be served with the provider's prefix attached, broken down by endpoint.",
    )
    aliases: Dict[Literal["links", "references", "structures"], Dict[str, str]] = Field(
        {},
        description="A mapping between field names in the database with their corresponding OPTIMADE field names, broken down by endpoint.",
    )

    length_aliases: Dict[
        Literal["links", "references", "structures"], Dict[str, str]
    ] = Field(
        {},
        description=(
            "A mapping between a list property (or otherwise) and an integer property that defines the length of that list, "
            "for example elements -> nelements. The standard aliases are applied first, so this dictionary must refer to the "
            "API fields, not the database fields."
        ),
    )

    index_links_path: Path = Field(
        Path(__file__).parent.joinpath("index_links.json"),
        description="Absolute path to a JSON file containing the MongoDB collection of /links resources for the index meta-database",
    )

    @root_validator(pre=True)
    def load_default_settings(cls, values):  # pylint: disable=no-self-argument
        """
        Loads settings from a root file if available and uses that as defaults in
        place of built in defaults
        """
        config_file_path = Path(values.get("config_file", DEFAULT_CONFIG_FILE_PATH))

        new_values = {}

        if config_file_path.exists() and config_file_path.is_file():
            logger.debug("Found config file at: %s", config_file_path)
            with open(config_file_path) as f:
                new_values = json.load(f)
        else:
            logger.debug(  # pragma: no cover
                "Did not find config file at: %s", config_file_path
            )

        new_values.update(values)

        return new_values

    class Config:
        env_prefix = "optimade_"


CONFIG = ServerConfig()



# pylint: disable=import-error
import pytest

from .utils import get_min_ver

min_ver = get_min_ver("pymatgen")
pymatgen = pytest.importorskip(
    "pymatgen",
    minversion=min_ver,
    reason=f"pymatgen must be installed with minimum version {min_ver} for these tests to"
    " be able to run",
)

from pymatgen import Molecule, Structure as PymatgenStructure

from optimade.adapters import Structure
from optimade.adapters.structures.pymatgen import (
    get_pymatgen,
    _get_structure,
    _get_molecule,
)


def test_successful_conversion(RAW_STRUCTURES):
    """Make sure its possible to convert"""
    for structure in RAW_STRUCTURES:
        assert isinstance(
            get_pymatgen(Structure(structure)), (PymatgenStructure, Molecule)
        )


def test_successful_conversion_structure(structure):
    """Make sure its possible to convert to pymatgen Structure"""
    assert isinstance(_get_structure(structure), PymatgenStructure)
    assert isinstance(get_pymatgen(structure), PymatgenStructure)


def test_null_lattice_vectors(null_lattice_vector_structure):
    """Make sure null lattice vectors are handled

    This also respresents a test for successful conversion to pymatgen Molecule
    """
    assert isinstance(_get_molecule(null_lattice_vector_structure), Molecule)
    assert isinstance(get_pymatgen(null_lattice_vector_structure), Molecule)


def test_null_positions(null_position_structure):
    """Make sure null positions are handled"""
    assert isinstance(get_pymatgen(null_position_structure), PymatgenStructure)
    assert isinstance(_get_structure(null_position_structure), PymatgenStructure)
    assert isinstance(_get_molecule(null_position_structure), Molecule)


def test_special_species(SPECIAL_SPECIES_STRUCTURES):
    """Make sure vacancies and non-chemical symbols ("X") are handled"""
    for special_structure in SPECIAL_SPECIES_STRUCTURES:
        structure = Structure(special_structure)
        assert isinstance(get_pymatgen(structure), PymatgenStructure)



from builtins import str
from builtins import zip
from builtins import range
from builtins import object
from functools import reduce
import json
import numpy as np
import sys
import logging
from nomadcore import parser_backend
from nomadcore.unit_conversion import unit_conversion

class CachingLevel(object):
    Forward = 1
    Cache = 2
    ForwardAndCache = 3
    Ignore = 4
    PreOpenedCache = 5
    PreOpenedIgnore = 6

    @classmethod
    def toText(cls, cl):
        if cl == cls.Forward:
            return "Forward"
        elif cl == cls.ForwardAndCache:
            return "ForwardAndCache"
        elif cl == cls.Cache:
            return "Cache"
        elif cl == cls.Ignore:
            return "Ignore"
        elif cl == cls.PreOpenedIgnore:
            return "PreOpenedIgnore"
        elif cl == cls.PreOpenedCache:
            return "PreOpenedIgnore"
        else:
            return "CachingLevel(%s)" % cl

    @classmethod
    def restrict(cls, a, b):
        """restricts a keeping in account that it is contained in b (i.e. if b is not forwarded then also a cannot be forwarded)"""
        if b == cls.Cache or b == cls.Ignore:
            if a == cls.Forward:
                return cls.Ignore
            elif a == cls.ForwardAndCache:
                return cls.Cache
        return a
class CachingSection(object):
    """Represents an open section, and can cache values"""
    def __init__(self, gIndex, references, storeInSuper = False, activeBackend=None):
        self.gIndex = gIndex
        self.references = references
        self.simpleValues = {}
        self.arrayValues = {}
        self.subSectionValues = {}
        self.storeInSuper = storeInSuper
        self.activeBackend = activeBackend

    def __getitem__(self, metaName):
        """Returns the cached values corresponding to metaName, or None if
        there aren't any. You can search values and subsections.
        """
        res = self.simpleValues.get(metaName, None)
        if res:
            return res
        res = self.arrayValues.get(metaName, None)
        if res:
            return res
        res = self.subSectionValues.get(metaName, None)
        return res

    def get_latest_value(self, metaname_list):
        """Goes through the given list of metanames and tries to find the data
        by succesfully opening the latest value from within this section and
        it's subsections.  If the given path leads to an actual value, it is
        returned.

        Args:
            metaname_list: list of metainfo names

        Returns:
            The value that if found when successively opening the data in this
            section and it's subsection. If some of the metanames in the given
            list are not found, returns None.
        """
        # Make path into a list
        if not isinstance(metaname_list, (tuple, list)):
            metaname_list = [metaname_list]

        # Go deeper into the section structure by iterating through the given
        # list of metanames.
        current = self
        for metaname in metaname_list:
            new_section = current[metaname]
            if new_section is not None:
                current = new_section[-1]
            else:
                return None

        return current

    def add_latest_value(self, metaname_list, metaname, require=False):
        latest_value = self.get_latest_value(metaname_list)
        if latest_value is not None:
            self.activeBackend.addValue(metaname, latest_value)
        else:
            if require:
                raise KeyError("Could not add the latest value from path '{}' because not value was found.")

    def add_latest_array_values(self, metaname_list, metaname, require=False):
        latest_value = self.get_latest_value(metaname_list)
        if latest_value is not None:
            self.activeBackend.addArrayValues(metaname, latest_value)
        else:
            if require:
                raise KeyError("Could not add the latest value from path '{}' because not value was found.")

    def addValue(self, metaInfo, value):
        vals = self.simpleValues.get(metaInfo.name, None)
        if vals is None:
            self.simpleValues[metaInfo.name] = [value]
        else:
            vals.append(value)

    def setArrayValues(self, metaInfo, values, offset = None):
        vals = self.arrayValues.get(metaInfo.name, None)
        if vals is None:
            raise Exception("setArrayValues(%s,...) called before adding a value" % metaInfo.name)
        else:
            if offset:
                idxs = [slice(offset[i], offset[i] + values.shape[i]) for i in range(len(offset))]
            else:
                idxs = [slice(0, x) for x in values.shape]
            vals[len(vals) - 1][idxs] = values

    def addArrayValues(self, metaInfo, values):
        vals = self.arrayValues.get(metaInfo.name, None)
        if vals is None:
            self.arrayValues[metaInfo.name] = [values]
        else:
            vals.append(values)

    def addSubsection(self, metaInfo, section):
        vals = self.subSectionValues.get(metaInfo.name, None)
        if vals is None:
            self.subSectionValues[metaInfo.name] = [section]
        else:
            vals.append(section)

    def __str__(self):
        return """CachingSection{{
        gIndex:{0},
        references:{1},
        simpleValues:{2},
        arrayValues.keys:{3},
        subSectionValues.keys:{4},
        storeInSuper:{5},
    }}""".format(
        self.gIndex,
        self.references,
        self.simpleValues,
        list(self.arrayValues.keys()),
        list(self.subSectionValues.keys()),
        self.storeInSuper)

class CachingSectionManager(object):

    def __init__(self, metaInfo, parentSectionNames, lastSectionGIndex = -1, openSections = {}, onClose = [], onOpen = [], storeInSuper = False, forwardOpenClose = True, preOpened = False):
        self.metaInfo = metaInfo
        self.parentSectionNames = parentSectionNames
        self.lastSectionGIndex = lastSectionGIndex
        self.openSections = dict(openSections)
        self.onClose = list(onClose)
        self.onOpen = list(onOpen)
        self.storeInSuper = storeInSuper
        self.forwardOpenClose = forwardOpenClose
        self.preOpened = preOpened

    def setSectionInfo(self, gIndex, references):
        self.openSections[gIndex].references = [references[x] for x in self.parentSectionNames]

    def get_latest_section(self):
        """Returns the latest opened section if it is still open.
        """
        section = self.openSections.get(self.lastSectionGIndex)
        return section

    def openSection(self, backend):
        newGIndex = self.lastSectionGIndex + 1
        self.openSectionWithGIndex(backend, newGIndex)
        return newGIndex

    def openSectionWithGIndex(self, backend, gIndex):
        self.lastSectionGIndex = gIndex
        references = []
        for parentName in self.parentSectionNames:
            pSect = backend.sectionManagers.get(parentName, None)
            if pSect:
                references.append(pSect.lastSectionGIndex)
            else:
                references.append(-1)
        opened = CachingSection(gIndex, references, storeInSuper = self.storeInSuper, activeBackend=backend)
        self.openSections[gIndex] = opened
        for onOpenF in self.onOpen:
            onOpenF(backend, gIndex, opened)

    def closeSection(self, backend, gIndex):
        toClose = self.openSections[gIndex]
        for onCloseF in self.onClose:
            onCloseF(backend, gIndex, toClose)
        if toClose.storeInSuper:
            for superGIndex, superName in zip(toClose.references,self.parentSectionNames):
                superSect = backend.sectionManagers[superName].openSections.get(superGIndex, None)
                if superSect:
                    superSect.addSubsection(self.metaInfo, toClose)
                else:
                    backend.storeToClosedSuper(superName, superGIndex, self.metaInfo, toClose)
            if not toClose.references: # checks for empty list
                backend.parsingSession.addSubsection(self.metaInfo, toClose)
        if self.forwardOpenClose and backend.superBackend:
            backend.superBackend.closeSection(self.metaInfo.name, gIndex)
        del self.openSections[gIndex]

    def openSectionInfo(self, gIndex):
        section = self.openSections.get(gIndex, None)
        if section:
            return "section {name} ({references})".format(
                name = section.name,
                references = list(zip(self.parentSectionNames, section.references)))
        else:
            return "section {gIndex} in {name} is not open!!".format(
                name = self.metaInfo.name,
                gIndex = gIndex)

    def addValue(self, valueMetaInfo, value, gIndex):
        if (gIndex == -1):
            gI = self.lastSectionGIndex
        else:
            gI = gIndex
        try:
            self.openSections[gI].addValue(valueMetaInfo, value)
        except:
            raise Exception("Cannot add value for metadata %s to section %d (%d) of %s, as it is not open" % (valueMetaInfo.name, gI, gIndex, self.metaInfo.name))

    def setArrayValues(self, valueMetaInfo, value, offset = None, gIndex = -1):
        if gIndex == -1:
            gI = self.lastSectionGIndex
        else:
            gI = gIndex
        try:
            self.openSections[gI].setArrayValues(valueMetaInfo, value, offset)
        except:
            raise Exception("Cannot set array values for metadata %s to section %d (%d) of %s, as it is not open" % (valueMetaInfo.name, gI, gIndex, self.metaInfo.name))

    def addArrayValues(self, valueMetaInfo, value, offset = None, gIndex = -1):
        if gIndex == -1:
            gI = self.lastSectionGIndex
        else:
            gI = gIndex
        try:
            self.openSections[gI].addArrayValues(valueMetaInfo, value)
        except:
            raise Exception("Cannot add array values for metadata %s to section %d (%d) of %s, as it is not open" % (valueMetaInfo.name, gI, gIndex, self.metaInfo.name))


class CachingDataManager(object):
    def __init__(self, metaInfo, superSectionManager, cachingLevel):
        self.metaInfo = metaInfo
        self.superSectionManager = superSectionManager
        self.cachingLevel = cachingLevel


class ActiveBackend(object):
    def __init__(
            self, metaInfoEnv, sectionManagers, dataManagers, superBackend,
            propagateStartFinishParsing=True, default_units=None, metainfo_units=None):

        self.__metaInfoEnv = metaInfoEnv
        self.sectionManagers = sectionManagers
        self.dataManagers = dataManagers
        self.superBackend = superBackend
        self.propagateStartFinishParsing = propagateStartFinishParsing
        self.default_units = default_units  # A mapping between dimension and an unit definition.
        self.metainfo_units = metainfo_units  # A mapping between metaname and an unit definition.

    @classmethod
    def activeBackend(
            cls,
            metaInfoEnv,
            cachingLevelForMetaName={},
            defaultDataCachingLevel=CachingLevel.ForwardAndCache,
            defaultSectionCachingLevel=CachingLevel.Forward,
            superBackend=None,
            onClose={}, onOpen={},
            propagateStartFinishParsing=True,
            default_units=None,
            metainfo_units=None):

        for sectionName in onClose.keys():
            if sectionName not in metaInfoEnv:
                raise Exception(
                    "Found trigger for non existing section %s" % sectionName)

            elif metaInfoEnv.infoKinds[sectionName].kindStr != "type_section":
                raise Exception(
                    "Found trigger for %s which is not a section but %s" %
                    (sectionName, json.dumps(metaInfoEnv.infoKinds[sectionName].toDict(), indent=2)))

        for sectionName in onOpen.keys():
            if sectionName not in metaInfoEnv:
                raise Exception(
                    "Found trigger for non existing section %s" % sectionName)

            elif metaInfoEnv.infoKinds[sectionName].kindStr != "type_section":
                raise Exception(
                    "Found trigger for %s which is not a section but %s" %
                    (sectionName, json.dumps(metaInfoEnv.infoKinds[sectionName].toDict(), indent=2)))

        sectionManagers = {}
        for ik in metaInfoEnv.infoKinds.values():
            if ik.kindStr == "type_section":
                parentS, parentO = list(
                    metaInfoEnv.firstAncestorsByType(ik.name).get("type_section", [[], []]))
                parentS.sort()

                cachingLevel = reduce(
                    CachingLevel.restrict,
                    [
                        cachingLevelForMetaName.get(x, defaultSectionCachingLevel)
                        for x in ([ik.name] + parentS + parentO)])

                sectionManagers[ik.name] = CachingSectionManager(
                    metaInfo=ik,
                    parentSectionNames=parentS,
                    storeInSuper=(cachingLevel == CachingLevel.ForwardAndCache or cachingLevel == CachingLevel.Cache or cachingLevel == CachingLevel.PreOpenedCache),
                    forwardOpenClose=(cachingLevel == CachingLevel.Forward or cachingLevel == CachingLevel.ForwardAndCache),
                    preOpened=(cachingLevel == CachingLevel.PreOpenedCache or cachingLevel == CachingLevel.PreOpenedIgnore),
                    onClose=onClose.get(ik.name, []),
                    onOpen=onOpen.get(ik.name, []))

        dataManagers = {}
        for ik in metaInfoEnv.infoKinds.values():
            if ik.kindStr == "type_document_content" or ik.kindStr == "type_dimension":
                superSectionNames = metaInfoEnv.firstAncestorsByType(ik.name).get("type_section", [[]])[0]
                if not superSectionNames:
                    raise Exception("MetaInfo of conrete value %s is not in any superSection" % ik.name)
                elif len(superSectionNames) > 1:
                    raise Exception("MetaInfo of concrete value %s has multiple superSections (%s)" %
                                    (ik.name, superSectionNames))
                sectionManager = sectionManagers[superSectionNames[0]]
                dataManagers[ik.name] = CachingDataManager(
                    ik, sectionManager,
                    CachingLevel.restrict(
                        cachingLevelForMetaName.get(ik.name, defaultDataCachingLevel),
                        CachingLevel.Forward if sectionManager.forwardOpenClose or sectionManager.preOpened else CachingLevel.Ignore))

        return ActiveBackend(
            metaInfoEnv, sectionManagers, dataManagers, superBackend,
            propagateStartFinishParsing, default_units, metainfo_units)

    def appendOnClose(self, sectionName, onClose):
        self.sectionManagers.onClose.append(onClose)

    def get_latest_section(self, metaname):
        """Returns the latest opened section with the given metaname if it has
        not been closed.
        """
        manager = self.sectionManagers.get(metaname)
        if manager:
            return manager.get_latest_section()

    def appendOnOpen(self, sectionName, onOpen):
        self.sectionManagers.onOpen.append(onOpen)

    def storeToClosedSuper(self, superName, superGIndex, metaInfo, toClose):
        logging.getLogger("nomadcore.caching_backend").warn(
            "Dropping section {name} gIndex {gIndex}, as it cannot be added to closed super section {superName} gIndex: {superGIndex}".format(
                name = metaInfo.name,
                gIndex = toClose.gIndex,
                superName = superName,
                superGIndex = superGIndex))

    def startedParsingSession(self, mainFileUri, parserInfo, parsingStatus = None, parsingErrors = None):
        """should be called when the parsing starts, parserInfo should be a valid json dictionary"""
        if self.propagateStartFinishParsing and self.superBackend:
            self.superBackend.startedParsingSession(mainFileUri, parserInfo, parsingStatus, parsingErrors)

    def finishedParsingSession(self, parsingStatus, parsingErrors, mainFileUri = None, parserInfo = None,
                               parsingStats = None):
        """should be called when the parsing finishes"""
        if self.propagateStartFinishParsing and self.superBackend:
            self.superBackend.finishedParsingSession(parsingStatus, parsingErrors, mainFileUri, parserInfo, parsingStats)

    def addMatchTelemetry(self, match_telemetry, gIndex = -1):
        """ should be called for outputting match telemetry data:
        input data, together with capture info """
        if self.superBackend:
            self.superBackend.addMatchTelemetry(match_telemetry, gIndex)

    def metaInfoEnv(self):
        """the metaInfoEnv this parser was optimized for"""
        return self.__metaInfoEnv

    def openSections(self):
        """returns the sections that are still open
        sections are identified by metaName and their gIndex"""
        res = set()
        for manager in self.sectionManagers:
            res.extend(map(lambda x: (manager.metaInfo.name, x), manager.openSections.keys()))
        return res

    def sectionInfo(self, metaName, gIndex):
        """returns information on a section (for debugging purposes)"""
        manager = self.sectionManagers[metaName]
        section = manager.openSections.get(gIndex, None)
        if section:
            return "open section {} gIndex: {}, references: {}".format(metaName, gIndex, section.references)
        else:
            return "closed section {} gIndex: {}".format(metaName, gIndex)

    def openSection(self, metaName):
        """opens a new section and returns its new unique gIndex"""
        manager = self.sectionManagers[metaName]
        if self.superBackend and manager.forwardOpenClose:
            newGIndex = self.superBackend.openSection(metaName)
            manager.openSectionWithGIndex(self, newGIndex)
        else:
            newGIndex = manager.openSection(self)
        return newGIndex

    def openNonOverlappingSection(self, metaName):
        """opens a new section that is expected not to overlap with itself and returns its new unique gIndex"""
        manager = self.sectionManagers[metaName]
        if len(manager.openSections) != 0:
            raise Exception("Section %s was not expected to overlap" % metaName)
        return self.openSection(metaName)

    def openSectionWithGIndex(self, metaName, gIndex):
        """opens a new section where gIndex is generated externally
        gIndex should be unique (no reopening of a closed section)"""
        manager = self.sectionManagers[metaName]
        if self.superBackend and manager.forwardOpenClose:
            self.superBackend.openSectionWithGIndex(metaName, newGIndex)
        manager.openSectionWithGIndex(newGIndex)

    def setSectionInfo(self, metaName, gIndex, references):
        """sets info values of an open section
        references should be a dictionary with the gIndexes of the root sections this section refers to"""
        manager = self.sectionManagers[metaName]
        manager.setSectionInfo(gIndex, references)
        if self.superBackend and manager.forwardOpenClose:
            self.superBackend.setSectionInfo(metaName, gIndex, references)

    def closeSection(self, metaName, gIndex):
        """closes a section
        after this no other value can be added to the section
        metaName is the name of the meta info, gIndex the index of the section"""
        manager = self.sectionManagers[metaName]
        manager.closeSection(self, gIndex)

    def closeNonOverlappingSection(self, metaName):
        """closes a non overlapping section
        after this no other value can be added to the section
        metaName is the name of the meta info, gIndex the index of the section"""
        manager = self.sectionManagers[metaName]
        openGIndexes = list(manager.openSections.keys())
        if len(openGIndexes) != 1:
            if not openGIndexes:
                raise Exception("Call to closeNonOverlapping(%s) with no open section" % metaName)
            else:
                raise Exception("Section %s was not supposed to overlap, found %s open when closing" % (metaName, openGIndexes))
        manager.closeSection(self, openGIndexes[0])

    def addValue(self, metaName, value, gIndex = -1):
        """adds a json value corresponding to metaName
        the value is added to the section the meta info metaName is in
        a gIndex of -1 means the latest section
        """
        dataManager = self.dataManagers[metaName]
        cachingLevel = dataManager.cachingLevel
        if cachingLevel == CachingLevel.Forward or cachingLevel == CachingLevel.ForwardAndCache:
            if self.superBackend:
                self.superBackend.addValue(metaName, value, gIndex)
        if cachingLevel == CachingLevel.ForwardAndCache or cachingLevel == CachingLevel.Cache or cachingLevel == CachingLevel.PreOpenedCache:
            dataManager.superSectionManager.addValue(dataManager.metaInfo, value, gIndex)

    def addRealValue(self, metaName, value, gIndex = -1, unit=None):
        """Adds a floating point value corresponding to metaName The value is
        added to the section the meta info metaName is in A gIndex of -1 means
        the latest section.
        """
        if unit is not None:
            value = self.convert_unit(metaName, value, unit)

        dataManager = self.dataManagers[metaName]
        cachingLevel = dataManager.cachingLevel
        if cachingLevel == CachingLevel.Forward or cachingLevel == CachingLevel.ForwardAndCache:
            if self.superBackend:
                self.superBackend.addRealValue(metaName, value, gIndex)
        if cachingLevel == CachingLevel.ForwardAndCache or cachingLevel == CachingLevel.Cache or cachingLevel == CachingLevel.PreOpenedCache:
            dataManager.superSectionManager.addValue(dataManager.metaInfo, value, gIndex)

    def addArray(self, metaName, shape, gIndex = -1):
        """adds a new array value of the given size corresponding to metaName
        the value is added to the section the meta info metaName is in
        a gIndex of -1 means the latest section
        the array is unitialized"""
        dataManager = self.dataManagers[metaName]
        cachingLevel = dataManager.cachingLevel
        if cachingLevel == CachingLevel.Forward or cachingLevel == CachingLevel.ForwardAndCache:
            if self.superBackend:
                self.superBackend.addArray(metaName, shape, gIndex)
        if cachingLevel == CachingLevel.ForwardAndCache or cachingLevel == CachingLevel.Cache or cachingLevel == CachingLevel.PreOpenedCache:
            dataManager.superSectionManager.addArray(dataManager.metaInfo, shape, gIndex)

    def setArrayValues(self, metaName, values, offset = None, gIndex = -1, unit=None):
        """Adds values to the last array added, array must be a numpy array
        """
        if unit is not None:
            values = self.convert_unit(metaName, values, unit)

        dataManager = self.dataManagers[metaName]
        cachingLevel = dataManager.cachingLevel
        if cachingLevel == CachingLevel.Forward or cachingLevel == CachingLevel.ForwardAndCache or cachingLevel == CachingLevel.PreOpenedCache:
            if self.superBackend:
                self.superBackend.setArrayValues(metaName, values, offset, gIndex)
        if cachingLevel == CachingLevel.ForwardAndCache or cachingLevel == CachingLevel.Cache or cachingLevel == CachingLevel.PreOpenedCache:
            dataManager.superSectionManager.setArrayValues(dataManager.metaInfo, values, offset, gIndex)

    def addArrayValues(self, metaName, values, gIndex = -1, unit=None):
        """Adds an array value with the given array values.  Values must be a
        numpy array.
        """
        if unit is not None:
            values = self.convert_unit(metaName, values, unit)

        dataManager = self.dataManagers[metaName]
        cachingLevel = dataManager.cachingLevel
        if cachingLevel == CachingLevel.Forward or cachingLevel == CachingLevel.ForwardAndCache:
            if self.superBackend:
                self.superBackend.addArrayValues(metaName, values, gIndex=gIndex)
        if cachingLevel == CachingLevel.ForwardAndCache or cachingLevel == CachingLevel.Cache or cachingLevel == CachingLevel.PreOpenedCache:
            dataManager.superSectionManager.addArrayValues(dataManager.metaInfo, values, gIndex=gIndex)

    def convertScalarStringValue(self, metaName, strValue):
        """converts a scalar string value of the given meta info to a python value"""
        metaInfo = self.metaInfoEnv().infoKindEl(metaName)
        dtypeStr = metaInfo.dtypeStr
        return parser_backend.valueForStrValue(strValue, dtypeStr)

    def arrayForMetaInfo(self, metaName, shape):
        """Returns an array with the correct type for the given meta info, and the given shape"""
        metaInfo = self.metaInfoEnv().infoKindEl(metaName)
        dtypeStr = metaInfo.dtypeStr
        return np.zeros(shape, dtype = parser_backend.numpyDtypeForDtypeStr(dtypeStr))

    def convert_unit(self, metaName, value, unit):
        """Try to perform a unit conversion.
        """
        # If a unit has been specified in the metainfo, check that the
        # conversion can be done
        metaInfo = self.metaInfoEnv().infoKindEl(metaName)
        metainfo_unit = metaInfo.units
        if metainfo_unit:
            metainfo_dim = unit_conversion.ureg(metainfo_unit).dimensionality
            source_dim = unit_conversion.ureg(unit).dimensionality
            if metainfo_dim != source_dim:
                raise Exception("Unit conversion error with metainfo '{}'. Could not convert the given unit '{}' to target unit '{}' specified in the metainfo.".format(metaName, unit, metainfo_unit))

        # If there is a metaname specific unit request, convert to it
        target_unit = None
        if self.metainfo_units is not None:
            target_unit = self.metainfo_units.get(metaName)

        # If there is a dimension specific unit request, convert to it
        if target_unit is None:
            if self.default_units is not None:
                source_dim = unit_conversion.ureg(unit).dimensionality
                map_unit = self.default_units.get(str(source_dim))
                if map_unit:
                    target_unit = map_unit

        converted = unit_conversion.convert_unit(value, unit, target_unit)
        if converted is None:
            raise Exception("ActiveBackend cannot convert to unit '{}'".format(unit))
        return converted

    def pwarn(self, msg):
        """Writes a parser warning message"""
        self.addValue("parsing_message_warning_run", msg)



# Copyright 2015-2018 Martina Stella, Massimo Riello, Fawzi Mohamed, Ankit Kariryaa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy as np
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
from nomadcore.unit_conversion.unit_conversion import convert_unit
import json, os, re

############################################################
# This file contains functions that are needed
# by more than one parser.
############################################################

def get_metaInfo(filePath):
    """Loads metadata.

    Args:
        filePath: Location of metadata.

    Returns:
        metadata which is an object of the class InfoKindEnv in nomadcore.local_meta_info.py.
    """
    metaInfoEnv, warnings = loadJsonFile(filePath = filePath, dependencyLoader = None, extraArgsHandling = InfoKindEl.ADD_EXTRA_ARGS, uri = None)
    return metaInfoEnv


# Copyright 2015-2018 Massimo Riello, Adam Fekete, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABCMeta, abstractproperty


# Class hierarchy in LAMMPS source code:
# LAMMPS
# - Memory
# - Error
# - Universe
# - Input
#    - Variable
#    - Command
#       - Finish
#       - Special
# - Atom
#    - AtomVec
# - Update
#    - Integrate
#    - Min
# - Neighbor
#    - NeighList
#    - NeighRequest
# - Comm
#    - Irregular
# - Domain
#    - Region
#    - Lattice
# - Force
#    - Pair
#    - Bond
#    - Angle
#    - Dihedral
#    - Improper
#    - KSpace
#       - FFT3D
#       - Remap
# - Modify
#    - Fix
#    - Compute
# - Group
# - Output
#    - Thermo
#    - Dump
#    - WriteRestart
# - Timer


# Build a class hierarchy for easier operator handling

class Lammps(metaclass=ABCMeta):
    """Abstract superclass of all Lammps related classes"""
    _row_format = '{:>15} {}'
    pass


class Data(Lammps):
    def __init__(self):
        from LAMMPS.Domain import Domain
        from LAMMPS.Atom import AtomAtomic
        self.domain = Domain()   # simulation box
        self.atom = AtomAtomic()

        self.system = None
        self.communication = None
        self.computes = None
        self.dumps = None
        self.fixes = None
        self.groups = None
        self.variables = None

    # def __str__(self):
        # return self._row_format.format('Units:', self.style)

    def info(self):
        out = (
            "\n"
            "LAMMPS object:\n"
            "Domain:\n==============\n{domain}\n").format(
            domain=self.domain)
        return out


class Lattice(Lammps):
    """Define a lattice for use by other commands. In LAMMPS, a lattice is simply a set of points in space, determined by a unit cell with basis atoms, that is replicated infinitely in all dimensions. The arguments of the lattice command can be used to define a wide variety of crystallographic lattices."""
    pass


class Region(Lammps):
    pass


class Group(Lammps):
    pass


class Neighbor(Lammps):
    """This command sets parameters that affect the building of pairwise neighbor lists. All atom pairs within a neighbor cutoff distance equal to the their force cutoff plus the skin distance are stored in the list. """
    pass


class Units(Lammps):
    @abstractproperty
    def style(self): pass

    @abstractproperty
    def mass(self): pass

    @abstractproperty
    def distance(self): pass

    @abstractproperty
    def time(self): pass

    @abstractproperty
    def energy(self): pass

    @abstractproperty
    def velocity(self): pass

    @abstractproperty
    def force(self): pass

    @abstractproperty
    def torque(self): pass

    @abstractproperty
    def temperature(self): pass

    @abstractproperty
    def pressure(self): pass

    @abstractproperty
    def dynamic_viscosity(self): pass

    @abstractproperty
    def charge(self): pass

    @abstractproperty
    def dipole(self): pass

    @abstractproperty
    def electric_field(self): pass

    @abstractproperty
    def density(self, dim): pass

    def __str__(self):
        return self._row_format.format('Units:', self.style)

    def info(self):
        out = (
            "\n"
            "Unit sytle:            {style} \n"
            "   mass:               {mass:.6e} \n"
            "   distance:           {distance:.6e} \n"
            "   time:               {time:.6e} \n"
            "   energy:             {energy:.6e} \n"
            "   velocity:           {velocity:.6e} \n"
            "   force:              {force:.6e} \n"
            "   torque:             {torque:.6e} \n"
            "   temperature:        {temperature:.6e} \n"
            "   pressure:           {pressure:.6e} \n"
            "   dynamic_viscosity:  {dynamic_viscosity:.6e} \n"
            "   charge:             {charge:.6e} \n"
            "   dipole:             {dipole:.6e} \n"
            "   electric_field:     {electric_field:.6e} \n"
            "   density:            {density:.6e} \n").format(
            style=self.style,
            mass=self.mass,
            distance=self.distance,
            time=self.time,
            energy=self.energy,
            velocity=self.velocity,
            force=self.force,
            torque=self.torque,
            temperature=self.temperature,
            pressure=self.pressure,
            dynamic_viscosity=self.dynamic_viscosity,
            charge=self.charge,
            dipole=self.dipole,
            electric_field=self.electric_field,
            density=self.density)
        return out


###################################################
# Force fields:
###################################################


class AtomStyle(Lammps):
    @abstractproperty
    def name(self): pass

    def __str__(self):
        return self._row_format.format('Atom style:', self.name)


class PairStyle(Lammps):
    @abstractproperty
    def name(self): pass

    def __str__(self):
        return self._row_format.format('Pair style:', self.name)


class BondStyle(Lammps):
    @abstractproperty
    def name(self): pass

    def __str__(self):
        return self._row_format.format('Bond style:', self.name)


class AngleStyle(Lammps):
    @abstractproperty
    def name(self): pass

    def __str__(self):
        return self._row_format.format('Angle style:', self.name)


class DihedralStyle(Lammps):
    @abstractproperty
    def name(self): pass

    def __str__(self):
        return self._row_format.format('Dihedral style:', self.name)


class ImproperStyle(Lammps):
    @abstractproperty
    def name(self): pass

    def __str__(self):
        return self._row_format.format('Improper style:', self.name)


class KSpaceStyle(Lammps):
    @abstractproperty
    def name(self): pass

    def __str__(self):
        return self._row_format.format('KSpace style:', self.name)


class SpetialBonds(Lammps):
    pass


###################################################
# Abstract class hierarchy:
###################################################

class Input(Lammps):
    """The Input class reads an input script, stores variables,
    and invokes stand- alone commands that are child classes
    of the Command class"""
    pass


class Variable(Input):
    """The Command class is a parent class for certain input script
    commands that perform a one-time operation before/after/between
    simulations or which invoke a simulation."""
    pass


class Command(Input):
    """The Command class is a parent class for certain input script
    commands that perform a one-time operation before/after/between
    simulations or which invoke a simulation."""
    pass


class Finish(Command):
    """The Finish class is instantiated to print statistics to the screen
    after a simulation is performed, by commands like run and minimize."""
    pass


class Special(Command):
    """The Special class walks the bond topology of a molecular system to
    find 1st, 2nd, 3rd neighbors of each atom. It is invoked by several commands,
    like read data, read restart, and replicate."""
    pass


class Atom(Lammps):
    """The Atom class stores all per-atom arrays.
    More precisely, they are allo- cated and stored by the AtomVec class,
    and the Atom class simply stores a pointer to them.
    The AtomVec class is a parent class for atom styles,
    defined by the atom style command."""
    pass


class Neighbor(Lammps):
    """The Neighbor class builds and stores neighbor lists. The NeighList class stores
    a single list (for all atoms). The NeighRequest class is called by pair, fix, or
    compute styles when they need a particular kind of neighbor list."""
    pass


class Domain(Lammps):
    """The Domain class stores the simulation box geometry, as well as geometric Regions
    and any user definition of a Lattice. The latter are defined by region and
    lattice commands in an input script."""
    pass


class Modify(Lammps):
    """The Modify class stores lists of Fix and Compute classes, both of which are parent styles."""
    pass


class Group(Lammps):
    """The Group class manipulates groups that atoms are assigned to via the group command.
    It also computes various attributes of groups of atoms."""
    pass


class Output(Lammps):
    """The Output class is used to generate 3 kinds of output from a LAMMPS simulation:
    thermodynamic information printed to the screen and log file, dump file snapshots, and restart files.
    These correspond to the Thermo, Dump, and WriteRestart classes respectively.
    The Dump class is a parent style."""
    pass


if __name__ == '__main__':
    lmp = Data()
    print(lmp)
    print(lmp.info())



# Copyright 2015-2018 Massimo Riello, Adam Fekete, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

class LammpsVariables:

    def __init__(self):
        self.vars = {}

    def eval(self, exp):
        pass

    def parse(self, input):
        """parse
        :param input: "variable name style args"
        :return:
        """
        line = input.split(maxsplit=3)
        try:
            getattr(self, line[1])(line[0], line[2:])
        except AttributeError:
            print(line[1] + ": currently unsupported")

            # width self.read() as name,style,args:

    def delete(self, name, *args):
        """delete
        delete = no args
        """
        del self.vars[name]

    def index(self, name, *args):
        """index
        index args = one or more strings
        """
        self.vars[name] = args

    def loop(self, name, *args):
        """loop
        loop args = N
          N = integer size of loop, loop from 1 to N inclusive
        loop args = N pad
          N = integer size of loop, loop from 1 to N inclusive
          pad = all values will be same length, e.g. 001, 002, ..., 100
        loop args = N1 N2
          N1,N2 = loop from N1 to N2 inclusive
        loop args = N1 N2 pad
          N1,N2 = loop from N1 to N2 inclusive
          pad = all values will be same length, e.g. 050, 051, ..., 100
        """
        self.vars[name] = args

    def world(self, name, *args):
        """word
        world args = one string for each partition of processors
        """
        self.vars[name] = args

    def universe(self, name, *args):
        """universe
        universe args = one or more strings
        """
        self.vars[name] = args

    def uloop(self, name, *args):
        """uloop
        uloop args = N
          N = integer size of loop
        uloop args = N pad
          N = integer size of loop
          pad = all values will be same length, e.g. 001, 002, ..., 100
        """
        self.vars[name] = args

    def string(self, name, *args):
        """String vaiable
        string arg = one string
        """
        self.vars[name] = args

    def format(self, name, *args):
        """format
        format args = vname fstr
          vname = name of equal-style variable to evaluate
          fstr = C-style format string
        """
        self.vars[name] = args

    def getenv(self, name, *args):
        """getenv
        getenv arg = one string
        """
        self.vars[name] = args

    def file(self, name, *args):
        """file
        file arg = filename
        """
        self.vars[name] = args

    def atomfile(self, name, *args):
        """atomfile
        atomfile arg = filename
        """
        self.vars[name] = args

    def python(self, name, *args):
        """python
        python arg = function
        """
        self.vars[name] = args

    """equal
    equal or vector or atom args = one formula containing numbers, thermo keywords, math operations, group functions, atom values and vectors, compute/fix/variable references
    numbers = 0.0, 100, -5.4, 2.8e-4, etc
    constants = PI, version, on, off, true, false, yes, no
    thermo keywords = vol, ke, press, etc from thermo_style
    math operators = (), -x, x+y, x-y, x*y, x/y, x^y, x%y,
    x == y, x != y, x < y, x <= y, x > y, x >= y, x && y, x || y, !x
    math functions = sqrt(x), exp(x), ln(x), log(x), abs(x),
    sin(x), cos(x), tan(x), asin(x), acos(x), atan(x), atan2(y,x),
    random(x,y,z), normal(x,y,z), ceil(x), floor(x), round(x)
    ramp(x,y), stagger(x,y), logfreq(x,y,z), logfreq2(x,y,z),
    stride(x,y,z), stride2(x,y,z,a,b,c),
    vdisplace(x,y), swiggle(x,y,z), cwiggle(x,y,z)
    group functions = count(group), mass(group), charge(group),
    xcm(group,dim), vcm(group,dim), fcm(group,dim),
    bound(group,dir), gyration(group), ke(group),
    angmom(group,dim), torque(group,dim),
    inertia(group,dimdim), omega(group,dim)
    region functions = count(group,region), mass(group,region), charge(group,region),
    xcm(group,dim,region), vcm(group,dim,region), fcm(group,dim,region),
    bound(group,dir,region), gyration(group,region), ke(group,reigon),
    angmom(group,dim,region), torque(group,dim,region),
    inertia(group,dimdim,region), omega(group,dim,region)
    special functions = sum(x), min(x), max(x), ave(x), trap(x), slope(x), gmask(x), rmask(x), grmask(x,y), next(x)
    feature functions = is_active(category,feature,exact), is_defined(category,id,exact)
    atom value = id[i], mass[i], type[i], mol[i], x[i], y[i], z[i], vx[i], vy[i], vz[i], fx[i], fy[i], fz[i], q[i]
    atom vector = id, mass, type, mol, x, y, z, vx, vy, vz, fx, fy, fz, q
    compute references = c_ID, c_ID[i], c_ID[i][j], C_ID, C_ID[i]
    fix references = f_ID, f_ID[i], f_ID[i][j], F_ID, F_ID[i]
    variable references = v_name, v_name[i]
    """


    def equal(self, name, *args):
        self.vars[name] = args

    def vector(self, name, *args):
        self.vars[name] = args

    def atom(self, name, *args):
        self.vars[name] = args







if __name__ == '__main__':

    vars = LammpsVariables()

    examples = [
        'variable x index run1 run2 run3 run4 run5 run6 run7 run8',
        'variable LoopVar loop $n',
        'variable beta equal temp/3.0',
        'variable b1 equal x[234]+0.5*vol',
        'variable b1 equal "x[234] + 0.5*vol"',
        'variable b equal xcm(mol1,x)/2.0',
        'variable b equal c_myTemp',
        'variable b atom x*y/vol',
        'variable foo string myfile',
        'variable myPy python increase',
        'variable f file values.txt',
        'variable temp world 300.0 310.0 320.0 ${Tfinal}',
        'variable x universe 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15',
        'variable x uloop 15 pad',
        'variable str format x %.6g',
        'variable x delete',
    ]

    for line in examples:
        exp = line[len('variable'):]
        print(exp)
        vars.parse(exp)

    pass





# Copyright 2016-2018 Lauri Himanen, Fawzi Mohamed, Ankit Kariryaa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import os
import pickle
import numpy as np
from nomadcore.baseclasses import AbstractBaseParser
from cpmdparser.generic.inputparsing import metainfo_data_prefix, metainfo_section_prefix


class CPMDInputParser(AbstractBaseParser):
    """Parses the CPMD input file.
    """
    def __init__(self, parser_context):
        """
        """
        super(CPMDInputParser, self).__init__(parser_context)
        self.input_tree = None

    def parse(self, filepath):
        self.setup_input_tree(self.parser_context.version_id)
        self.collect_input(filepath)
        self.analyze_input()
        self.fill_metadata()

    def collect_input(self, filepath):
        """This function will first go through the input file and gather the
        information to the input tree.

        The data is not directly pushed to the backend inside this function
        because it is safer to first validate the whole structure and only then
        push it.
        """
        # The input file should be relatively small so we are better off
        # loading it into memory all at once.
        lines = None
        with open(filepath, "r") as fin:
            lines = fin.readlines()

        # Read the input line by line
        section_stack = []
        section_name = None
        section_object = None
        old_keyword_object = None
        parameters = []
        input_tree = self.input_tree
        input_tree.accessed = True

        for line in lines:
            line = line.strip()
            keyword_object = None

            # Skip empty lines
            if len(line) == 0:
                continue
            # Section ends
            if line.upper().startswith('&END'):
                if parameters:
                    if old_keyword_object is not None:
                        old_keyword_object.parameters = "\n".join(parameters)
                    else:
                        section_object.default_keyword = "\n".join(parameters)
                    parameters = []
                old_keyword_object = None
                section_stack.pop()
            # Section starts
            elif line[0] == '&':
                section_name = line[1:]
                section_stack.append(section_name)
                section_object = input_tree.get_section(section_name)
                section_object.accessed = True
            # Keywords, parameters
            else:
                # Try to find a keyword object
                splitted = line.split()
                current_keyword_name = []
                i_match = None
                for i_part, part in enumerate(splitted):
                    current_keyword_name.append(part)
                    current_keyword_object = section_object.get_keyword(" ".join(current_keyword_name))
                    if current_keyword_object is not None:
                        keyword_object = current_keyword_object
                        i_match = i_part

                # If multiple keywords with this name exist, choose the correct
                # one based on switch
                if isinstance(keyword_object, list):
                    switch = splitted[i_match]
                    correct_keyword_object = None
                    for keyword in keyword_object:
                        if switch in keyword.available_options:
                            correct_keyword_object = keyword
                            break
                    if correct_keyword_object is None:
                        raise LookupError("Could not find the correct keyword for '{}' in the input structure.".format(line))
                    keyword_object = correct_keyword_object

                # If keyword object found, place the options and save any
                # parameters that were found before
                if keyword_object is not None:
                    # print(keyword_object.name)
                    if parameters:
                        if old_keyword_object is not None:
                            old_keyword_object.parameters = "\n".join(parameters)
                            parameters = []
                    options = splitted[i_match+1:]
                    if options:
                        options = " ".join(options)
                    else:
                        options = None
                    keyword_object.options = options
                    keyword_object.accessed = True
                    old_keyword_object = keyword_object

                # If no keyword was found, and a section is open, the line is a
                # parameter line
                if keyword_object is None and len(section_stack) != 0:
                    parameters.append(line)

    def analyze_input(self):

        # Get the trajectory print settings
        root = self.input_tree
        cpmd = root.get_section("CPMD")
        trajectory = cpmd.get_keyword("TRAJECTORY")
        if trajectory:
            options = trajectory.options
            if options:
                if "RANGE" in options:
                    self.cache_service["trajectory_range"] = True
                if "SAMPLE" in options:
                    self.cache_service["trajectory_sample"] = True
                    parameters = trajectory.parameters
                    try:
                        lines = parameters.split("\n")
                        print_freq = int(lines[-1])
                    except (IndexError, ValueError):
                        self.cache_service["print_freq"] = None
                    else:
                        self.cache_service["print_freq"] = print_freq

        # Get the periodicity settings
        system = root.get_section("SYSTEM")
        symmetry = system.get_keyword("SYMMETRY")
        symmetry_parameters = symmetry.parameters.strip()
        cluster = system.get_keyword("CLUSTER")
        surface = system.get_keyword("SURFACE")
        polymer = system.get_keyword("POLYMER")

        # Bulk
        if symmetry_parameters != "0" and symmetry_parameters != "ISOLATED" and cluster.accessed is False:
            periodicity = [True, True, True]

        # Surface
        elif surface.accessed:
            options = surface.options
            if options is None:
                options = "XY"
            else:
                options = options.strip()
            if options == "XY":
                periodicity = [True, True, False]
            elif options == "XZ":
                periodicity = [True, False, True]
            elif options == "YZ":
                periodicity = [False, True, True]

        # Polymer
        elif polymer.accessed:
            periodicity = [True, False, False]

        # Isolated
        elif cluster.accessed or symmetry_parameters == "ISOLATED" or symmetry_parameters == "0":
            periodicity = [False, False, False]

        self.cache_service["configuration_periodic_dimensions"] = np.array(periodicity)

        # Get the XC functional
        class XCFunctional(object):
            def __init__(self, name, weight=1, parameters=None):
                self.name = name
                self.weight = weight
                self.parameters = parameters

        xc_list = []
        dft = root.get_section("DFT")
        if dft is not None:
            functional = dft.get_keyword("FUNCTIONAL")
            if functional is not None:
                xc = functional.options
                if xc is not None:
                    xc = xc.strip()

                    if xc == "LDA":
                        xc_list.append(XCFunctional("LDA_XC_TETER93"))

                    elif xc == "BLYP":
                        xc_list.append(XCFunctional("GGA_X_B88"))
                        xc_list.append(XCFunctional("GGA_C_LYP"))

                    elif xc == "B3LYP":
                        xc_list.append(XCFunctional("HYB_GGA_XC_B3LYP"))

                    elif xc == "PBE":
                        xc_list.append(XCFunctional("GGA_X_PBE"))
                        xc_list.append(XCFunctional("GGA_C_PBE"))

                    elif xc == "OLYP":
                        xc_list.append(XCFunctional("GGA_X_OPTX"))
                        xc_list.append(XCFunctional("GGA_C_LYP"))

                    elif xc == "HCTH":
                        xc_list.append(XCFunctional("GGA_XC_HCTH_120"))

                    elif xc == "PBE0":
                        xc_list.append(XCFunctional("HYB_GGA_XC_PBEH"))

                    elif xc == "BP":
                        xc_list.append(XCFunctional("GGA_X_B88"))
                        xc_list.append(XCFunctional("GGA_C_P86"))

                    elif xc == "XLYP":
                        xc_list.append(XCFunctional("GGA_XC_XLYP"))

                    elif xc == "PBES":
                        xc_list.append(XCFunctional("GGA_C_PBE_SOL"))
                        xc_list.append(XCFunctional("GGA_X_PBE_SOL"))

                    elif xc == "REVPBE":
                        xc_list.append(XCFunctional("GGA_C_PBE"))
                        xc_list.append(XCFunctional("GGA_X_PBE_R"))

                    elif xc == "TPSS":
                        xc_list.append(XCFunctional("MGGA_C_TPSS"))
                        xc_list.append(XCFunctional("MGGA_X_TPSS"))

                    # This version of OPTX is not yet found on the official list
                    # elif xc == "OPTX":

                    elif xc == "B1LYP":
                        xc_list.append(XCFunctional("HYB_GGA_XC_B1LYP"))

                    elif xc == "X3LYP":
                        xc_list.append(XCFunctional("HYB_GGA_XC_X3LYP"))

                    elif xc == "HSE06":
                        xc_list.append(XCFunctional("HYB_GGA_XC_HSE06"))

        # Sort the functionals alphabetically by name
        xc_list.sort(key=lambda x: x.name)
        xc_summary = ""

        # For every defined functional, stream the information to the
        # backend and construct the summary string
        for i, functional in enumerate(xc_list):

            gId = self.backend.openSection("section_XC_functionals")
            self.backend.addValue("XC_functional_name", functional.name)
            self.backend.addValue("XC_functional_weight", functional.weight)
            if functional.parameters is not None:
                pass
            self.backend.closeSection("section_XC_functionals", gId)

            if i != 0:
                xc_summary += "+"
            xc_summary += "{}*{}".format(functional.weight, functional.name)
            if functional.parameters is not None:
                xc_summary += ":{}".format()

        # Stream summary
        if xc_summary is not "":
            self.backend.addValue("XC_functional", xc_summary)

    def fill_metadata(self):
        """Goes through the input data and pushes everything to the
        backend.
        """
        def fill_metadata_recursively(section):
            """Recursively goes through the input sections and pushes everything to the
            backend.
            """
            if not section.accessed:
                return

            if section.name != "":
                section_name = metainfo_section_prefix + "{}".format(section.name)
            else:
                section_name = metainfo_section_prefix[:-1]
            gid = self.backend.openSection(section_name)

            # Keywords
            for keyword_list in section.keywords.values():
                for keyword in keyword_list:
                    if keyword.accessed:
                        # Open keyword section
                        keyword_name = metainfo_section_prefix + "{}.{}".format(section.name, keyword.unique_name.replace(" ", "_"))
                        key_id = self.backend.openSection(keyword_name)

                        # Add options
                        if keyword.options:
                            option_name = metainfo_data_prefix + "{}.{}_options".format(section.name, keyword.unique_name.replace(" ", "_"))
                            self.backend.addValue(option_name, keyword.options)

                        # Add parameters
                        if keyword.parameters:
                            parameter_name = metainfo_data_prefix + "{}.{}_parameters".format(section.name, keyword.unique_name.replace(" ", "_"))
                            self.backend.addValue(parameter_name, keyword.parameters)

                        # Close keyword section
                        self.backend.closeSection(keyword_name, key_id)

            # # Default keyword
            default_keyword = section.default_keyword
            if default_keyword is not None:
                name = metainfo_data_prefix + "{}_default_keyword".format(section.name)
                self.backend.addValue(name, default_keyword)

            # Subsections
            for subsection in section.subsections.values():
                fill_metadata_recursively(subsection)

            self.backend.closeSection(section_name, gid)

        fill_metadata_recursively(self.input_tree)

    def setup_input_tree(self, version_number):
        """Loads the version specific pickle file which contains pregenerated
        input structure.
        """
        pickle_path = os.path.dirname(__file__) + "/input_data/cpmd_input_tree.pickle"
        input_tree_pickle_file = open(pickle_path, 'rb')
        self.input_tree = pickle.load(input_tree_pickle_file)



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from flask_restplus import Resource, abort
from flask import request

from nomad.datamodel import OptimadeEntry

from .api import api
from .common import base_request_args, base_search_request, nentries, ns
from .models import json_api_single_response_model, entry_listing_endpoint_parser, Meta, \
    Links as LinksModel, single_entry_endpoint_parser, base_endpoint_parser, \
    json_api_info_response_model, json_api_list_response_model, EntryDataObject, \
    get_entry_properties, to_calc_with_metadata
from .filterparser import parse_filter, FilterException


@ns.route('/structures')
class StructureList(Resource):
    @api.doc('structures')
    @api.response(400, 'Invalid requests, e.g. bad parameter.')
    @api.expect(entry_listing_endpoint_parser, validate=True)
    @api.marshal_with(json_api_list_response_model, skip_none=True, code=200)
    def get(self):
        ''' Returns a list of structures that match the given optimade filter expression. '''
        response_fields = base_request_args()

        try:
            filter = request.args.get('filter', None)
            page_limit = int(request.args.get('page_limit', 10))
            page_number = int(request.args.get('page_number', 1))
            sort = request.args.get('sort', 'chemical_formula_reduced')

        except Exception:
            abort(400, message='bad parameter types')  # TODO Specific json API error handling

        search_request = base_search_request().include('calc_id', 'upload_id')

        if filter is not None:
            try:
                search_request.query(parse_filter(filter))
            except FilterException as e:
                abort(400, message='Could not parse filter expression: %s' % str(e))

        result = search_request.execute_paginated(
            page=page_number,
            per_page=page_limit)
        # order_by='optimade.%s' % sort)  # TODO map the Optimade property

        returned = result['pagination']['total']
        results = to_calc_with_metadata(result['results'])
        assert len(results) == len(result['results']), 'archive and elasticsearch are not consistent'

        return dict(
            meta=Meta(
                query=request.url,
                returned=returned,
                available=nentries(),
                last_id=results[-1].calc_id if returned > 0 else None),
            links=LinksModel(
                'structures',
                returned=returned,
                page_number=page_number,
                page_limit=page_limit,
                sort=sort, filter=filter),
            data=[EntryDataObject(d, optimade_type='structures', response_fields=response_fields) for d in results]
        ), 200


@ns.route('/structures/<string:id>')
class Structure(Resource):
    @api.doc('structure')
    @api.response(400, 'Invalid requests, e.g. bad parameter.')
    @api.response(404, 'Id does not exist.')
    @api.expect(single_entry_endpoint_parser, validate=True)
    @api.marshal_with(json_api_single_response_model, skip_none=True, code=200)
    def get(self, id: str):
        ''' Retrieve a single structure for the given id '''
        response_fields = base_request_args()
        search_request = base_search_request().search_parameters(calc_id=id)

        result = search_request.execute_paginated(
            page=1,
            per_page=1)

        available = result['pagination']['total']
        results = to_calc_with_metadata(result['results'])
        assert len(results) == len(result['results']), 'Mongodb and elasticsearch are not consistent'

        if available == 0:
            abort(404, 'The calculation with id %s does not exist' % id)

        return dict(
            meta=Meta(query=request.url, returned=1),
            data=EntryDataObject(results[0], optimade_type='structures', response_fields=response_fields)
        ), 200


@ns.route('/info/structures')
class StructuresInfo(Resource):
    @api.doc('structures_info')
    @api.response(400, 'Invalid requests, e.g. bad parameter.')
    @api.expect(base_endpoint_parser, validate=True)
    @api.marshal_with(json_api_info_response_model, skip_none=True, code=200)
    def get(self):
        ''' Returns information about the structures endpoint implementation '''
        base_request_args()

        result = {
            'description': 'a structure entry',
            'properties': get_entry_properties(),
            'formats': ['json'],
            'output_fields_by_format': {
                'json': list(OptimadeEntry.m_def.all_properties.keys())}
        }

        return dict(
            meta=Meta(query=request.url, returned=1),
            data=result
        ), 200



# Copyright 2016-2018 Ask Hjorth Larsen, Fawzi Mohamed
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

#
# Main author and maintainer: Ask Hjorth Larsen <asklarsen@gmail.com>

from __future__ import print_function
import os
import sys
import logging
from glob import glob
import re

import numpy as np
from ase.data import chemical_symbols
from ase import Atoms

from nomadcore.simple_parser import (mainFunction, SimpleMatcher as SM,
                                     AncillaryParser)
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
from nomadcore.unit_conversion.unit_conversion \
    import register_userdefined_quantity, convert_unit

from siestaparser.inputvars import varlist

parser_info = {'name':'siesta-parser', 'version': '1.0'}


def siesta_energy(title, meta, **kwargs):
    return SM(r'siesta:\s*%s\s*=\s*(?P<%s__eV>\S+)' % (title, meta),
              name=meta, **kwargs)


def line_iter(fd, linepattern = re.compile(r'\s*([^#]+)')):
    # Strip off comments and whitespace, return only non-empty strings
    for line in fd:
        match = linepattern.match(line)
        if match:
            line = match.group().rstrip()
            if line:
                yield line


def get_input_metadata(inputvars_file, use_new_format):
    inputvars = {}
    blocks = {}

    varset = set(varlist)

    lower_vars = {}
    for var in varlist:
        lower_vars[var.lower()] = var

    def addvar(tokens):
        name = tokens[0]
        val = ' '.join(tokens[1:])
        name = name.lower()
        if name in lower_vars:
            name = lower_vars[name]
            inputvars[name] = val

    currentblock = None

    with open(inputvars_file) as fd:
        lines = line_iter(fd)

        for line in lines:
            tokens = line.split()
            assert len(tokens) > 0

            if tokens[0].startswith('%'):
                if tokens[0].lower() == '%block':
                    #assert currentblock == None
                    currentblock = []
                    blocks[tokens[1]] = currentblock
                elif tokens[0].lower() == '%endblock':
                    currentblock = None
                else:
                    raise ValueError('Unknown: %s' % tokens[0])
            else:
                if use_new_format:
                    if line.startswith(' '):
                        if currentblock is None:
                            continue  # Ignore.  Probably some warning
                        #assert currentblock is not None, line
                        currentblock.append(tokens)
                    else:
                        currentblock = None
                        addvar(tokens)
                else:
                    if currentblock is not None:
                        currentblock.append(tokens)
                    else:
                        addvar(tokens)

    return inputvars, blocks


"""
%block PAO.Basis                 # Define Basis set
O                     2                    # Species label, number of l-shells
 n=2   0   2                         # n, l, Nzeta
   3.305      2.479
   1.000      1.000
 n=2   1   2 P   1                   # n, l, Nzeta, Polarization, NzetaPol
   3.937      2.542
   1.000      1.000
H                     1                    # Species label, number of l-shells
 n=1   0   2 P   1                   # n, l, Nzeta, Polarization, NzetaPol
   4.709      3.760
   1.000      1.000
%endblock PAO.Basis
"""


class SiestaContext(object):
    def __init__(self):
        self.reset()

    def reset(self):
        self.fname = None  # The file that we are parsing
        self.dirname = None  # Base directory of calculations
        #self.parser = None  # The parser object
        self.format = None  # 'old' or 'new'; when parsing version

        # label and things determined by label
        self.label = None
        self.files = None  # Dict of files
        self.blocks = None  # Dict of input blocks (coords, cell, etc.)

        self.data = {}
        self.special_input_vars = {}

        self.system_meta = {}
        self.section_refs = {}  # {name: gindex, ...}
        self.simulation_type = None

    def adhoc_format_new(self, parser):
        assert self.format is None
        self.format = 'new'

    def adhoc_format_old(self, parser):
        assert self.format is None
        self.format = 'old'

    def adhoc_set_label(self, parser):
        # ASSUMPTION: the parser fIn is in the 'root' of whatever was uploaded.
        # This may not be true.  Figure out how to do this in general.
        line = parser.fIn.readline()
        assert line.startswith('reinit: System Label:')
        self.label = label = line.split()[-1]
        dirname = self.dirname

        files = {}

        for fileid in ['EIG', 'KP']:
            path = os.path.join(dirname, '%s.%s' % (label, fileid))
            if os.path.isfile(path):
                files[fileid] = path
            # Warn if files are not present?
            #
            # Also: input file
            #       input parser logfile
            #
            # what else?  We already get force/stress/positions from stdout.
        if self.format == 'new':
            inplogfiles = glob('%s/fdf-*.log' % dirname)
            assert len(inplogfiles) == 1
            if inplogfiles:
                inplogfiles.sort()
                files['inputlog'] = inplogfiles[0]
        else:
            assert self.format == 'old', self.format
            files['inputlog'] = os.path.join(dirname, 'out.fdf')
        self.files = files

    def adhoc_set_simulation_type(self, parser):
        line = parser.fIn.readline()

        if self.simulation_type is not None:
            return

        line = line.strip()

        if line.startswith('Single-point'):
            self.simulation_type = 'singlepoint'
        elif 'opt' in line or 'move' in line:
            self.simulation_type = 'optimization'
        else:
            raise ValueError('Todo: recognize simulation type "%s"' % line)

    def startedParsing(self, fname, parser):
        self.fname = fname
        path = os.path.abspath(fname)
        self.dirname = os.path.dirname(path)

    #def onClose_x_siesta_section_xc_authors(self, backend, gindex, section):

    def onClose_section_frame_sequence(self, backend, gindex, section):
        backend.addValue('frame_sequence_to_sampling_ref',
                         self.section_refs['sampling_method'])

    def onOpen_section_sampling_method(self, backend, gindex, section):
        self.section_refs['sampling_method'] = gindex

    def onOpen_section_frame_sequence(self, backend, gindex, section):
        self.section_refs['frame_sequence'] = gindex

    def onClose_section_sampling_method(self, backend, gindex, section):
        simtype = self.simulation_type
        assert simtype is not None
        if simtype == 'optimization':
            backend.addValue('sampling_method', 'geometry_optimization')
        elif simtype == 'singlepoint':
            pass
        else:
            raise ValueError('XXX: %s' % simtype)

    def onClose_section_eigenvalues(self, backend, gindex, section):
        self.read_eigenvalues(backend)

    def onOpen_section_method(self, backend, gindex, section):
        self.section_refs['method'] = gindex

    def onClose_section_method(self, backend, gindex, section):
        temp = self.special_input_vars['ElectronicTemperature']
        temp, unit = temp.split()
        assert unit == 'Ry'  # Siesta always converts to Ry here I think
        temp = float(temp)
        temp = convert_unit(temp, 'rydberg')
        backend.addValue('smearing_width', temp)

        #simtype = self.special_input_vars['MD.TypeOfRun']
        #print('SIMTYPE', simtype)
        #sdfsdf

    def onOpen_section_system(self, backend, gindex, section):
        self.section_refs['system'] = gindex

    def onClose_section_system(self, backend, gindex, section):
        data = self.data
        meta = self.system_meta

        latvec = data.pop('block_lattice_vectors', None)
        if latvec is not None:
            latvec = latvec.astype(float)
            size = self.special_input_vars['LatticeConstant']
            size, unit = size.split()
            #assert unit == 'ang', unit
            unit = {'ang': 'angstrom',
                    'Bohr': 'bohr'}[unit]

            size = float(size)
            size = convert_unit(size, unit)
            meta['simulation_cell'] = latvec * size

        cell = data.pop('outcell_ang', None)
        cell2 = data.pop('auto_unit_cell_ang', None)
        if cell2 is not None:
            cell = cell2

        if cell is not None:
            cell = cell.astype(float)
            cell = convert_unit(cell, 'angstrom')
            meta['simulation_cell'] = cell

        labels = data.pop('block_species_label', None)
        if labels is not None:
            assert labels.shape[1] == 1
            labels = labels[:, 0]
            self.labels = labels

        block_coords_and_species = self.data.pop('block_coords_and_species_from_inputlog', None)

        block_coords_and_species = data.pop('block_coords_and_species', block_coords_and_species)
        coords_and_species = data.pop('coords_and_species', None)

        if coords_and_species is None:
            coords_and_species = block_coords_and_species

        if coords_and_species is not None:
            coords = coords_and_species[:, :3].astype(float)

            unit = self.special_input_vars['AtomicCoordinatesFormat']
            if unit == 'Ang':
                coords = convert_unit(coords, 'angstrom')
            elif unit in ['Fractional', 'ScaledCartesian']:
                a = Atoms('%dX' % len(coords),
                          scaled_positions=coords,
                          cell=meta['simulation_cell'])
                coords = a.positions
            else:
                raise ValueError('Unknown: %s' % unit)
            meta['atom_positions'] = coords

            species_index = coords_and_species[:, 3].astype(int)

            atom_labels = np.array([self.labels[i - 1] for i in species_index])
            meta['atom_labels'] = atom_labels

        positions = self.data.pop('outcoord_ang', None)
        if positions is not None:
            positions = convert_unit(positions.astype(float), 'angstrom')
            meta['atom_positions'] = positions

        for key, value in meta.items():
            backend.addArrayValues(key, value)

        backend.addArrayValues('configuration_periodic_dimensions',
                               np.ones(3, bool))

        assert len(self.data) == 0, self.data

    def onClose_section_run(self, backend, gindex, section):
        pass

    def onClose_section_single_configuration_calculation(self, backend,
                                                         gindex, section):
        forces = self.data.pop('forces_ev_ang', None)
        if forces is not None:
            forces = forces.astype(float)
            forces = convert_unit(forces, 'eV/angstrom')
            backend.addArrayValues('atom_forces_free_raw', forces)

        stress = self.data.pop('stress_tensor_ev_ang', None)
        if stress is not None:
            stress = stress.astype(float)
            stress = convert_unit(stress, 'eV/angstrom^3')
            backend.addArrayValues('stress_tensor', stress)

        backend.addValue('single_configuration_to_calculation_method_ref',
                         self.section_refs['method'])
        backend.addValue('single_configuration_calculation_to_system_ref',
                         self.section_refs['system'])

    def onClose_x_siesta_section_input(self, backend, gindex, section):
        inputvars_file = self.files.get('inputlog')
        if inputvars_file is None:
            raise ValueError('no input logfile!')

        inputvars, blocks = get_input_metadata(inputvars_file,
                                               self.format == 'new')
        for varname, value in inputvars.items():
            backend.addValue('x_siesta_input_%s' % varname, value)

        for special_name in ['LatticeConstant',
                             'AtomicCoordinatesFormat',
                             'AtomicCoordinatesFormatOut',
                             'ElectronicTemperature']:
            self.special_input_vars[special_name] = inputvars.get(special_name)

        self.blocks = blocks
        self.data['block_coords_and_species_from_inputlog'] = np.array(blocks['AtomicCoordinatesAndAtomicSpecies'], str)

        authors = section['x_siesta_xc_authors']
        if authors is None:
            raise ValueError('XC authors not found!')

        assert len(authors) == 1
        authors = authors[0]

        # XXX Case sensitive?
        mapping = {'CA': ('LDA_X', 'LDA_C_PZ'),
                   'PZ': ('LDA_X', 'LDA_C_PZ'),
                   'PW92': ('LDA_X', 'LDA_C_PW'),
                   #'PW91': '',
                   'PBE': ('GGA_X_PBE', 'GGA_C_PBE'),
                   'revPBE': ('GGA_X_PBE_R', 'GGA_C_PBE'),
                   'RPBE': ('GGA_X_RPBE', 'GGA_C_PBE'),
                   #'WC': ('GGA_X_WC', ),
                   # Siesta does not mention which correlation is used with
                   # the WC functional.  Is it just the PBE one?
                   'AM05': ('GGA_X_AM05', 'GGA_C_AM05'),
                   'PBEsol': ('GGA_X_PBE_SOL', 'GGA_C_PBE_SOL'),
                   'BLYP': ('GGA_X_B88', 'GGA_C_LYP'),
                   'DF1': ('gga_x_pbe_r', 'vdw_c_df1'),
                   'DRSLL': ('gga_x_pbe_r', 'vdw_c_df1'),
                   'LMKLL': ('gga_x_rpw86', 'vdw_c_df2'),
                   'DF2': ('gga_x_rpw86', 'vdw_c_df2'),
                   'KBM': ('GGA_X_OPTB88_VDW', 'vdw_c_df1'),
                   'C09': ('GGA_X_C09X', 'vdw_c_df1'),
                   'BH': ('GGA_X_LV_RPW86', 'vdw_c_df1'),
        }
        xc = mapping.get(authors)

        if xc is None:
            raise ValueError('XC functional %s unsupported by parser'
                             % authors)

        for funcname in xc:
            gid = backend.openSection('section_XC_functionals')
            backend.addValue('XC_functional_name', funcname)
            backend.closeSection('section_XC_functionals', gid)

    def read_eigenvalues(self, backend):
        eigfile = self.files.get('EIG')
        if eigfile is None:
            return

        with open(eigfile) as fd:
            eFermi = float(next(fd).strip())
            nbands, nspins, nkpts = [int(n) for n in next(fd).split()]

            tokens = []
            for line in fd:
                tokens.extend(line.split())

        tokens = iter(tokens)

        eps = np.empty((nspins, nkpts, nbands))
        for k in range(nkpts):
            kindex = int(next(tokens))
            assert k + 1 == kindex
            for s in range(nspins):
                eps[s, k, :] = [next(tokens) for _ in range(nbands)]
        unread = list(tokens)
        assert len(unread) == 0
        assert s == nspins - 1
        assert k == nkpts - 1

        # Where does SIESTA store the occupations?
        backend.addArrayValues('eigenvalues_values', convert_unit(eps, 'eV'))

        kpfile = self.files.get('KP')
        if kpfile is None:
            return


        with open(kpfile) as fd:
            tokens = fd.read().split()

        nkpts = int(tokens[0])
        data = np.array(tokens[1:], object).reshape(-1, 5)
        coords = data[:, 1:4].astype(float)
        weights = data[:, 4].astype(float)
        backend.addArrayValues('eigenvalues_kpoints', coords)
        # XXX metadata for Fermi level?
        # k-point weights?

    def save_array(self, key, dtype=float, istart=0, iend=None,
                   unit=None):
        return get_array(key, dtype=dtype, istart=istart, iend=iend, unit=unit,
                         storage=self.data)

    def multi_sm(self, name, startpattern, linepattern, endmatcher=None,
                 conflict='fail',  # 'fail', 'keep', 'overwrite'
                 *args, **kwargs):

        pat = re.compile(linepattern)  # XXX how to get compiled pattern?
        ngroups = pat.groups

        allgroups = []
        def addline(parser):
            line = parser.fIn.readline()
            match = pat.match(line)
            assert match is not None
            thislinegroups = match.groups()
            assert len(thislinegroups) == ngroups
            allgroups.append(thislinegroups)

        def savearray(parser):
            arr = np.array(allgroups, dtype=object)
            del allgroups[:]
            if name in self.data:
                if conflict == 'fail':
                    raise ValueError('grrr %s %s' % (name, self.data[name]))
                elif conflict == 'keep':
                    return  # Do not save array
                elif conflict == 'overwrite':
                    pass
                else:
                    raise ValueError('Unknown keyword %s' % conflict)
            if arr.size > 0:
                self.data[name] = arr

        if endmatcher is None:
            endmatcher = r'.*'

        if hasattr(endmatcher, 'swapcase'):
            endmatcher = SM(endmatcher,
                            endReStr='',
                            forwardMatch=True,
                            name='%s-end' % name,
                            adHoc=savearray)

        sm = SM(startpattern,
                name=name,
                subMatchers=[
                    SM(linepattern,
                       name='%s-line' % name,
                       repeats=True,
                       forwardMatch=True,
                       required=True,
                       adHoc=addline),
                    endmatcher,
                ], **kwargs)
        return sm

context = SiestaContext()

def get_header_matcher():
    m = SM(r'',
           name='header',
           fixedStartValues={'program_name': 'Siesta',
                             'program_basis_set_type': 'numeric AOs'},
           weak=True,
           forwardMatch=True,
           subMatchers=[
               SM(r'Siesta Version: (?P<program_version>\S+)',
                  name='version',
                  adHoc=context.adhoc_format_new),
               SM(r'SIESTA\s*(?P<program_version>.+)',
                  name='alt-version', adHoc=context.adhoc_format_old),
               SM(r'Architecture\s*:\s*(?P<x_siesta_arch>.+)', name='arch'),
               SM(r'Compiler flags\s*:\s*(?P<x_siesta_compilerflags>.+)',
                  name='flags'),
           ])
    return m


def anycase(string):
    tokens = []
    for letter in list(string):
        if letter.isalpha():
            tokens.append('[%s%s]' % (letter.upper(),
                                      letter.lower()))
        else:
            tokens.append(letter)
    return ''.join(tokens)

welcome_pattern = r'\s*\*\s*WELCOME TO SIESTA\s*\*'

def get_input_matcher():
    m = SM(welcome_pattern,
           name='welcome',
           sections=['section_method', 'x_siesta_section_input'],
           fixedStartValues={'electronic_structure_method': 'DFT',
                             'smearing_kind': 'fermi'},
           subFlags=SM.SubFlags.Unordered,
           subMatchers=[
               SM(r'NumberOfAtoms\s*(?P<number_of_atoms>\d+)',
                  name='natoms'),
               context.multi_sm('block_species_label',
                                anycase(r'%block ChemicalSpeciesLabel'),
                                r'\s*\d+\s*\d+\s*(\S+)',
                            conflict='keep'),
               context.multi_sm('block_coords_and_species',
                                anycase(r'%block AtomicCoordinatesAndAtomicSpecies'),
                                r'\s*(\S+)\s*(\S+)\s*(\S+)\s*(\d+)'),
               context.multi_sm('block_lattice_vectors',
                                anycase(r'%block LatticeVectors'),
                                r'(?!%)\s*(\S+)\s*(\S+)\s*(\S+)'),
               SM(r'%s\s*(?P<x_siesta_xc_authors>\S+)' % anycase('xc\.authors'),
                  name='xc authors',
                  fixedStartValues={'x_siesta_xc_authors': 'CA'}),
               #SM(r'MD.TypeOfRun\s*(?P<x_siesta_typeofrun>\S+)',
               #   fixedStartValues={'x_siesta_typeofrun': 'none'}),
               SM(r'reinit: System Name:\s*(?P<system_name>.+)',
                  name='sysname'),
               SM(r'reinit: System Label:\s*(?P<x_siesta_system_label>\S+)',
                  name='syslabel', forwardMatch=True,
                  adHoc=context.adhoc_set_label),
               context.multi_sm('coords_and_species',
                                r'siesta: Atomic coordinates \(Bohr\) and species',
                                r'siesta:\s*(\S+)\s*(\S+)\s*(\S+)\s*(\d+)'),
               context.multi_sm('auto_unit_cell_ang',
                                r'siesta: Automatic unit cell vectors \(Ang\):',
                                r'siesta:\s*(\S+)\s*(\S+)\s*(\S+)'),
               # XXX must be an array with number of spin channels!
               #SM(r'Total number of electrons:\s*(?P<number_of_electrons>\S+)',
               #   name='nelectrons'),
       ])
    return m


step_pattern = r'\s*(Single-point calculation|Begin[^=]+=\s*\d+)'

def get_step_matcher():
    m = SM(step_pattern,
           name='step',
           forwardMatch=True,
           adHoc=context.adhoc_set_simulation_type,
           sections=['section_single_configuration_calculation'],
           subFlags=SM.SubFlags.Unordered,
           subMatchers=[
               SM(r'\s*=============+', name='====='),
               context.multi_sm('outcoord_ang',
                                r'outcoor: Atomic coordinates \(Ang\):',
                                r'\s*(\S+)\s*(\S+)\s*(\S+)'),
               context.multi_sm('outcell_ang',
                                r'outcell: Unit cell vectors \(Ang\):',
                                r'\s*(\S+)\s*(\S+)\s*(\S+)'),
               SM(r'siesta: E_KS\(eV\)\s*=\s*(?P<energy_total__eV>\S+)',
                  name='etotal'),
               context.multi_sm('forces_ev_ang',
                                r'siesta: Atomic forces \(eV/Ang\):',
                                r'(?:siesta:)?\s+\d+\s*(\S+)\s*(\S+)\s*(\S+)'),
               context.multi_sm('stress_tensor_ev_ang',
                                r'siesta: Stress tensor \(static\) \(eV/Ang\*\*3\):',
                                r'siesta:\s*(\S+)\s*(\S+)\s*(\S+)'),
               SM(r'siesta: Final energy \(eV\):',
                  endReStr='siesta:\s*Total\s*=\s*\S+',
                  name='Efinal',
                  subMatchers=[
                      siesta_energy('Band Struct\.', 'energy_sum_eigenvalues'),
                      siesta_energy('Kinetic', 'electronic_kinetic_energy'),
                      siesta_energy('Hartree', 'energy_electrostatic'),
                      #siesta_energy('Ext\. field', ''),
                      siesta_energy('Exch\.-corr\.', 'energy_XC'),
                      #siesta_energy('Ion-electron', ''),
                      #siesta_energy('Ion-Ion', ''),
                      #siesta_energy('Ekinion', ''),
                      # energy_total was matched above already
                      #siesta_energy('Total', 'energy_total'),
                      SM(r'', weak=True, name='trigger_readeig',
                         sections=['section_eigenvalues']),
                  ]),
           ])
    return m


mainFileDescription = SM(
    r'',
    name='root',
    #weak=True,
    sections=['section_run'],
    subMatchers=[
        get_header_matcher(),
        SM(r'(%s|%s)' % (welcome_pattern, step_pattern),
           name='system-section',
           #weak=True,
           forwardMatch=True,
           repeats=True,
           required=True,
           sections=['section_system', 'section_frame_sequence', 'section_sampling_method'],
           subMatchers=[
               get_input_matcher(),
               get_step_matcher(),
           ]),
        SM(r'x^',  # Make sure whole file is parsed
           name='end')
    ])


class SiestaParser():
   """ A proper class envolop for running this parser from within python. """
   def __init__(self, backend, **kwargs):
       self.backend_factory = backend

   def parse(self, mainfile):
       from unittest.mock import patch
       logging.info('siesta parser started')
       logging.getLogger('nomadcore').setLevel(logging.WARNING)
       backend = self.backend_factory("siesta.nomadmetainfo.json")
       context.reset()
       with patch.object(sys, 'argv', ['<exe>', '--uri', 'nmd://uri', mainfile]):
           mainFunction(
               mainFileDescription,
               None,
               parser_info,
               cachingLevelForMetaName={},
               superContext=context,
               superBackend=backend)

       return backend



# Copyright 2015-2018 Massimo Riello, Adam Fekete, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABCMeta, abstractproperty
from collections import OrderedDict
import numpy as np

from LAMMPS.BaseClasses import AtomStyle

from enum import Enum, unique
import unittest


def read_atoms(atom_style, ):
    pass


class Angle(AtomStyle):
    name = 'angle'
    keys = ['atom_id', 'molecule_id', 'atom_type', 'x', 'y', 'z']

class Atomic(AtomStyle):
    name = 'atomic'
    keys = ['atom_id', 'atom_type', 'x', 'y', 'z']

    def __init__(self, n_atoms = 0):
        self.n_atoms = n_atoms
        self.data = dict()



    def add_atom(self, atom_type, x, y, z):
        pass


    def __str__(self):
        row_format ='{:>15} {}'

        ret_str = '\n'.join([
            "Atoms",
            row_format.format('atom_style:', self.name),
            row_format.format('# of atoms:', self.n_atoms)
        ])
        return ret_str

class Body(AtomStyle):
    name = 'body'
    keys = ['atom_id', 'atom_type', 'bodyflag', 'mass', 'x', 'y', 'z']

class Bond(AtomStyle):
    name = 'bond'
    keys = ['atom_id', 'molecule_id', 'atom_type', 'x', 'y', 'z']

class Charge(AtomStyle):
    name = 'charge'
    keys = ['atom_id', 'atom_type', 'q', 'x', 'y', 'z']

class Dipole(AtomStyle):
    name = 'dipole'
    keys = ['atom_id', 'atom_type', 'q', 'x', 'y', 'z', 'mux', 'muy', 'muz']

class Dpd(AtomStyle):
    name = 'dpd'
    keys = ['atom_id', 'atom_type', 'theta', 'x', 'y', 'z']

class Electron(AtomStyle):
    name = 'electron'
    keys = ['atom_id', 'atom_type', 'q', 'spin', 'eradius', 'x', 'y', 'z']

class Ellipsoid(AtomStyle):
    name = 'ellipsoid'
    keys = ['atom_id', 'atom_type', 'ellipsoidflag', 'density', 'x', 'y', 'z']

class Full(AtomStyle):
    name = 'full'
    keys = ['atom_id', 'molecule_id', 'atom_type', 'q', 'x', 'y', 'z']

class Line(AtomStyle):
    name = 'line'
    keys = ['atom_id', 'molecule_id', 'atom_type', 'lineflag', 'density', 'x', 'y', 'z']

class Meso(AtomStyle):
    name = 'meso'
    keys = ['atom_id', 'atom_type', 'rho', 'e', 'cv', 'x', 'y', 'z']

class Molecular(AtomStyle):
    name = 'molecular'
    keys = ['atom_id', 'molecule_id', 'atom_type', 'x', 'y', 'z']

class Peri(AtomStyle):
    name = 'peri'
    keys = ['atom_id', 'atom_type', 'volume', 'density', 'x', 'y', 'z']

class Smd(AtomStyle):
    name = 'smd'
    keys = ['atom_id', 'atom_type', 'molecule', 'volume', 'mass', 'kernel-radius', 'contact-radius', 'x', 'y', 'z']

class Sphere(AtomStyle):
    name = 'sphere'
    keys = ['atom_id', 'atom_type', 'diameter', 'density', 'x', 'y', 'z']

class Template(AtomStyle):
    name = 'template'
    keys = ['atom_id', 'molecule_id', 'template-index', 'template-atom', 'atom_type', 'x', 'y', 'z']

class Tri(AtomStyle):
    name = 'tri'
    keys = ['atom_id', 'molecule_id', 'atom_type', 'triangleflag', 'density', 'x', 'y', 'z']

class Wavepacket(AtomStyle):
    name = 'wavepacket'
    keys = ['atom_id', 'atom_type', 'charge', 'spin', 'eradius', 'etag', 'cs_re', 'cs_im', 'x', 'y', 'z']

class Hybrid(AtomStyle):
    name = 'hybrid'
    keys = ['atom_id', 'atom_type', 'x', 'y', 'z']







def atom_style(style, *args):
    """Wrapper for atom_style command"""

    if style == 'body':
        return body(*args)

    elif style == 'template':
        return template(*args)

    elif style == 'hybrid':

        list = hybrid()
        args_iter = args.__iter__()

        while True:
            try:
                style = args_iter.__next__()

                if style == 'body':
                    bstyle = args_iter.__next__()
                    nmin = args_iter.__next__()
                    nmax = args_iter.__next__()
                    list.append(body(bstyle, nmin, nmax))

                elif style == 'template':
                    templat_id = args_iter.__next__()
                    list.append(template(templat_id))

                else:
                    list.append(eval(style)())

            except StopIteration:
                break
        return list
    else:
        return eval(style)()


    # # return eval(style)(*args)
    # cls = globals()[style]
    # return cls(*args)

    # # create an itarator
    # args_iter = args.__iter__()
    # while True:
    #     try:
    #         style = args_iter.__next__()
    #
    #
    #     except StopIteration:
    #         break
    #


class _atom_style(metaclass=ABCMeta):
    """Abstract class for atom_style"""
    pass

class _body_style(metaclass=ABCMeta): pass
class angle(_atom_style): pass
class atomic(_atom_style): pass

class body(_atom_style):
    def __init__(self, bstyle, Nmin, Nmax):
        self.bstyle = bstyle
        self.nmin = Nmin
        self.nmax = Nmax

class bond(_atom_style): pass
class charge(_atom_style): pass
class dipole(_atom_style): pass
class dpd(_atom_style): pass
class electron(_atom_style): pass
class ellipsoid(_atom_style): pass
class full(_atom_style): pass
class line(_atom_style): pass
class meso(_atom_style): pass
class molecular(_atom_style): pass
class peri(_atom_style): pass
class smd(_atom_style): pass
class sphere(_atom_style): pass
class tri(_atom_style): pass
# class wavepacket(_atom_style): pass
# not defined yet

class template(_atom_style):
    def __init__(self, id):
        self.id = id


class hybrid(_atom_style, list):
    pass
    # def __init__(self, *args):
    #
    #     # self.list = []
    #
    #     # create an itarator
    #     args_iter = args.__iter__()
    #     while True:
    #         try:
    #             style = args_iter.__next__()
    #
    #             if style == 'body':
    #                 bstyle = args_iter.__next__()
    #                 nmin = args_iter.__next__()
    #                 nmax = args_iter.__next__()
    #                 self.list.append(body(bstyle, nmin, nmax))
    #             elif style == 'template':
    #                 templat_id = args_iter.__next__()
    #                 self.list.append(template(templat_id))
    #             else:
    #                 self.list.append(eval(style))
    #
    #         except StopIteration:
    #             break




# class _Atoms(metaclass=ABCMeta):
#     """
#
#     """
#     pass
#
# @unique
# class _atom_styles(Enum):
#     angle = 1
#     atomic = 2
#     body = 3
#     bond = 4
#     charge = 5
#     dipole = 6
#     dpd = 7
#     electron = 8
#     ellipsoid = 9
#     full = 10
#     line = 11
#     meso = 12
#     molecular = 13
#     peri = 14
#     smd = 15
#     sphere = 16
#     tri = 17
#     template = 18
#     wavepacket = 19
#     hybrid = 20
#
#
# class atom_style():
#     # STYLES = ['angle', 'atomic', 'body', 'bond', 'charge', 'dipole', 'dpd', 'electron', 'ellipsoid', 'full', 'line', 'meso', 'molecular', 'peri', 'smd', 'sphere', 'template', 'tri', 'wavepacket']
#
#     def __init__(self, style, *args):
#         if style not in _atom_styles.__members__:
#             raise NameError('Wrong style type!')
#
#         self.style = _atom_styles[style]
#
#         if style is 'body':
#             self.bstyle = args[0]
#             self.bstyle_args = args[1:]
#
#         self.args = args
#         # for arg in args:
#         #     print(arg)
#
#
#
#
# class s(_Atoms):
#     def __init__(self):
#         self.style = ''




if __name__ == '__main__':

    a = np.array([1,2,3])
    print(a)


    a = Atomic(10)
    print(a)


    pass

    # atom_style('atomic')
    # atom_style('bond')
    # atom_style('full')
    # a = atom_style('body', 'nparticle', 2, 10)
    # print(a.bstyle)
    # print(a.nmin)
    # print(a.nmax)
    #
    # atom_style('hybrid', 'charge', 'bond')
    # a = atom_style('hybrid', 'charge', 'body', 'nparticle', 2, 5)
    # print(a.list)
    # atom_style('template', 'myMols')
